#
#  ktool | kmacho
#  structs.py
#
#  Custom Struct implementation reflecting behavior of named tuples while also handling behind-the-scenes packing/unpacking.
#
#  This file is part of ktool. ktool is free software that
#  is made available under the MIT license. Consult the
#  file "LICENSE" that is distributed together with this file
#  for the exact licensing terms.
#
#  Copyright (c) kat 2021.
#

type_mask = 0xffff0000
size_mask = 0xffff

type_uint = 0
type_sint = 0x10000
type_str = 0x20000
type_bytes = 0x30000

uint8_t = 1
uint16_t = 2
uint32_t = 4
uint64_t = 8

int8_t = type_sint | 1
int16_t = type_sint | 2
int32_t = type_sint | 4
int64_t = type_sint | 8


char_t = [type_str | i for i in range(65)]


def _bytes_to_hex(data) -> str:
    return data.hex()


def _uint_to_int(uint, bits):
    """
    Assume an int was read from binary as an unsigned int,

    decode it as a two's compliment signed integer

    :param uint:
    :param bits:
    :return:
    """
    if (uint & (1 << (bits - 1))) != 0:  # if sign bit is set e.g., 8bit: 128-255
        uint = uint - (1 << bits)  # compute negative value
    return uint  # return positive value as is


# noinspection PyUnresolvedReferences
class Struct:
    """
    Custom namedtuple-esque Struct representation. Can be unpacked from bytes or manually created with existing
        field values

    Subclassed and passed `fields` and `sizes` values.

    Fields are exposed as read-write attributes, and when written to, will update the backend
        byte representation of the struct, accessible via the .raw attribute

    """

    # noinspection PyProtectedMember
    @staticmethod
    def create_with_bytes(struct_class, raw, byte_order="little"):
        """
        Unpack a struct from raw bytes

        :param struct_class: Struct subclass
        :param raw: Bytes
        :param byte_order: Little/Big Endian Struct Unpacking
        :return: struct_class Instance
        """
        instance: Struct = struct_class(byte_order)
        current_off = 0
        raw = bytearray(raw)
        inst_raw = bytearray()

        # I *Genuinely* cannot figure out where in the program the size mismatch is happening. This should hotfix?
        raw = raw[:struct_class.SIZE]

        for field in instance._fields:
            value = instance._field_sizes[field]

            field_value = None

            if isinstance(value, int):
                field_type = type_mask & value
                size = size_mask & value

                data = raw[current_off:current_off + size]

                if field_type == type_str:
                    field_value = data.decode('utf-8').replace('\x00', '')

                elif field_type == type_bytes:
                    field_value = bytes(data)

                elif field_type == type_uint:
                    field_value = int.from_bytes(data, byte_order)

                elif field_type == type_sint:
                    field_value = int.from_bytes(data, byte_order)
                    field_value = _uint_to_int(field_value, size * 8)

            elif issubclass(value, Struct):
                size = value.SIZE
                data = raw[current_off:current_off+size]
                field_value = Struct.create_with_bytes(value, data)

            setattr(instance, field, field_value)
            inst_raw += data
            current_off += size

        instance.raw = inst_raw

        instance.pre_init()
        instance.initialized = True
        instance.post_init()

        return instance

    @staticmethod
    def create_with_values(struct_class, values, byte_order="little"):
        """
        Pack/Create a struct given field values

        :param byte_order:
        :param struct_class: Struct subclass
        :param values: List of values
        :return: struct_class Instance
        """

        instance: Struct = struct_class(byte_order)

        # noinspection PyProtectedMember
        for i, field in enumerate(instance._fields):
            instance.__setattr__(field, values[i])

        instance._rebuild_raw()
        instance.pre_init()
        instance.initialized = True
        instance.post_init()
        return instance

    def typename(self):
        return self.__class__.__name__

    def desc(self):
        return ""

    def __getattribute__(self, item):
        if item == 'raw':
            return self._rebuild_raw()
        return super().__getattribute__(item)

    def __str__(self):
        text = f'{self.__class__.__name__}('
        for field in self._fields:
            field_item = None
            if isinstance(self.__getattribute__(field), str):
                field_item = self.__getattribute__(field)
            elif isinstance(self.__getattribute__(field), bytearray) or isinstance(self.__getattribute__(field), bytes):
                field_item = self.__getattribute__(field)
            elif isinstance(self.__getattribute__(field), int):
                field_item = hex(self.__getattribute__(field))
            elif issubclass(self.__getattribute__(field), Struct):
                field_item = str(self.__getattribute__(field))
            text += f'{field}={field_item}, '
        return text[:-2] + ')'

    def render_indented(self, indent_size=2) -> str:
        text = f'{self.__class__.__name__}\n'
        for field in self._fields:
            field_item = None
            if isinstance(self.__getattribute__(field), str):
                field_item = self.__getattribute__(field)
            elif isinstance(self.__getattribute__(field), bytearray) or isinstance(self.__getattribute__(field), bytes):
                field_item = self.__getattribute__(field)
            elif isinstance(self.__getattribute__(field), int):
                field_item = hex(self.__getattribute__(field))
            elif issubclass(self.__getattribute__(field).__class__, Struct):
                field_item = '\n' + " "*(indent_size+2) + self.__getattribute__(field).render_indented(indent_size+2)
            text += f'{" "*indent_size}{field}={field_item}\n'
        return text

    def serialize(self):
        struct_dict = {
            'type': self.__class__.__name__
        }

        for field in self._fields:
            field_item = None
            if isinstance(self.__getattribute__(field), str):
                field_item = self.__getattribute__(field)
            elif isinstance(self.__getattribute__(field), bytearray) or isinstance(self.__getattribute__(field), bytes):
                field_item = _bytes_to_hex(self.__getattribute__(field))
            elif isinstance(self.__getattribute__(field), int):
                field_item = self.__getattribute__(field)
            elif issubclass(self.__getattribute__(field).__class__, Struct):
                field_item = self.__getattribute__(field).serialize()
            struct_dict[field] = field_item

        return struct_dict

    def __init__(self, fields=None, sizes=None, byte_order="little", no_patch=False):

        if sizes is None:
            raise AssertionError(
                "Do not use the bare Struct class; it must be implemented in an actual type; Missing Sizes")

        if fields is None:
            raise AssertionError(
                "Do not use the bare Struct class; it must be implemented in an actual type; Missing Fields")

        self.initialized = False
        self.no_patch = False

        self.super = super()
        self._fields = fields
        self.byte_order = byte_order

        self._field_sizes = {}

        for index, i in enumerate(fields):
            self._field_sizes[i] = sizes[index]

        self.off = 0

    def pre_init(self):
        """stub for subclasses. gets called before patch code is enabled"""
        pass

    def post_init(self):
        """stub for subclasses. gets called *after* patch code is enabled"""
        pass

    def _rebuild_raw(self):
        raw = bytearray()
        for field in self._fields:
            size = self._field_sizes[field]

            field_dat = self.__getattribute__(field)

            data = None

            if isinstance(field_dat, int):
                data = field_dat.to_bytes(size, byteorder=self.byte_order)
            elif isinstance(field_dat, bytearray) or isinstance(field_dat, bytes):
                data = field_dat
            elif isinstance(field_dat, str):
                data = field_dat.encode('utf-8')

            assert data is not None

            raw += bytearray(data)

        return raw


class BitFieldStruct(Struct):

    def __str__(self):
        text = f'{self.__class__.__name__}('
        for field in self._BITFIELDS:
            field_item = self.__getattribute__(field) if isinstance(self.__getattribute__(field), bytearray) else hex(
                self.__getattribute__(field))
            text += f'{field}={field_item}, '
        return text[:-2] + ')'

    def pre_init(self):

        assert sum(self._BF_SIZES) == self.SIZE * 8

        value = self.value

        cur = 0
        for i, bf_name in enumerate(self._BITFIELDS):
            bf_size = self._BF_SIZES[i]

            mask = int('1' * bf_size, 2)

            self.super.__setattr__(bf_name, (value >> cur) & mask)
            # print(f'{bf_name} = {value} >> {cur} & {mask}')
            cur += bf_size


class fat_header(Struct):
    """
    First 8 Bytes of a FAT MachO File

    Attributes:
        self.magic: FAT MachO Magic

        self.nfat_archs: Number of Fat Arch entries after these bytes
    """
    _FIELDNAMES = ['magic', 'nfat_archs']
    _SIZES = [uint32_t, uint32_t]
    SIZE = 8

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)


class fat_arch(Struct):
    """
    Struct representing a slice in a FAT MachO

    Attribs:
        cpu_type:
    """
    _FIELDNAMES = ['cpu_type', 'cpu_subtype', 'offset', 'size', 'align']
    _SIZES = [uint32_t,uint32_t,uint32_t,uint32_t,uint32_t]
    SIZE = 20

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)
        self.cpu_type = 0
        self.cpu_subtype = 0
        self.offset = 0
        self.size = 0
        self.align = 0


class mach_header(Struct):
    _FIELDNAMES = ['magic', 'cpu_type', 'cpu_subtype', 'filetype', 'loadcnt', 'loadsize', 'flags']
    _SIZES = [uint32_t, uint32_t, uint32_t, uint32_t, uint32_t, uint32_t, uint32_t]
    SIZE = 28

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)
        self.magic = 0
        self.cpu_type = 0
        self.cpu_subtype = 0
        self.filetype = 0
        self.loadcnt = 0
        self.loadsize = 0
        self.flags = 0


class mach_header_64(Struct):
    _FIELDNAMES = ['magic', 'cpu_type', 'cpu_subtype', 'filetype', 'loadcnt', 'loadsize', 'flags', 'reserved']
    _SIZES = [uint32_t, uint32_t, uint32_t, uint32_t, uint32_t, uint32_t, uint32_t, uint32_t]
    SIZE = 32

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)
        self.magic = 0
        self.cpu_type = 0
        self.cpu_subtype = 0
        self.filetype = 0
        self.loadcnt = 0
        self.loadsize = 0
        self.flags = 0
        self.reserved = 0


class unk_command(Struct):
    _FIELDNAMES = ['cmd', 'cmdsize']
    _SIZES = [uint32_t, uint32_t]
    SIZE = 8

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)
        self.cmd = 0
        self.cmdsize = 0


class segment_command(Struct):
    _FIELDNAMES = ['cmd', 'cmdsize', 'segname', 'vmaddr', 'vmsize', 'fileoff', 'filesize', 'maxprot',
                   'initprot', 'nsects', 'flags']
    _SIZES = [uint32_t, uint32_t, char_t[16], uint32_t, uint32_t, uint32_t, uint32_t, uint32_t, uint32_t, uint32_t, uint32_t]
    SIZE = 56

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)
        self.cmd = 0
        self.cmdsize = 0
        self.segname = 0
        self.vmaddr = 0
        self.vmsize = 0
        self.fileoff = 0
        self.filesize = 0
        self.maxprot = 0
        self.initprot = 0
        self.nsects = 0
        self.flags = 0


class segment_command_64(Struct):
    _FIELDNAMES = ['cmd', 'cmdsize', 'segname', 'vmaddr', 'vmsize', 'fileoff', 'filesize', 'maxprot',
                   'initprot', 'nsects', 'flags']
    _SIZES = [uint32_t, uint32_t, char_t[16], uint64_t, uint64_t, uint64_t, uint64_t, uint32_t, uint32_t, uint32_t, uint32_t]
    SIZE = 72

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)
        self.cmd = 0
        self.cmdsize = 0
        self.segname = 0
        self.vmaddr = 0
        self.vmsize = 0
        self.fileoff = 0
        self.filesize = 0
        self.maxprot = 0
        self.initprot = 0
        self.nsects = 0
        self.flags = 0


class section(Struct):
    _FIELDNAMES = ['sectname', 'segname', 'addr', 'size', 'offset', 'align', 'reloff',
                   'nreloc', 'flags', 'reserved1', 'reserved2']
    _SIZES = [char_t[16], char_t[16], uint32_t, uint32_t, uint32_t, uint32_t, uint32_t, uint32_t, uint32_t, uint32_t, uint32_t]
    SIZE = 68

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)
        self.sectname = 0
        self.segname = 0
        self.addr = 0
        self.size = 0
        self.offset = 0
        self.align = 0
        self.reloff = 0
        self.nreloc = 0
        self.flags = 0
        self.reserved1 = 0
        self.reserved2 = 0


class section_64(Struct):
    _FIELDNAMES = ['sectname', 'segname', 'addr', 'size', 'offset', 'align', 'reloff',
                   'nreloc', 'flags', 'reserved1', 'reserved2', 'reserved3']
    _SIZES = [char_t[16], char_t[16], uint64_t, uint64_t, uint32_t, uint32_t, uint32_t, uint32_t, uint32_t, uint32_t, uint32_t, uint32_t]
    SIZE = 32 + 16 + (4 * 8)

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)
        self.sectname = 0
        self.segname = 0
        self.addr = 0
        self.size = 0
        self.offset = 0
        self.align = 0
        self.reloff = 0
        self.nreloc = 0
        self.flags = 0
        self.reserved1 = 0
        self.reserved2 = 0
        self.reserved3 = 0


class symtab_command(Struct):
    _FIELDNAMES = ['cmd', 'cmdsize', 'symoff', 'nsyms', 'stroff', 'strsize']
    _SIZES = [uint32_t, uint32_t, uint32_t, uint32_t, uint32_t, uint32_t]
    SIZE = 4 * 6

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)
        self.cmd = 0
        self.cmdsize = 0
        self.symoff = 0
        self.nsyms = 0
        self.stroff = 0
        self.strsize = 0


class dysymtab_command(Struct):
    _FIELDNAMES = ['cmd', 'cmdsize', 'ilocalsym', 'nlocalsym', 'iextdefsym', 'nextdefsym',
                   'iundefsym', 'nundefsym', 'tocoff', 'ntoc', 'modtaboff', 'nmodtab',
                   'extrefsymoff', 'nextrefsyms', 'indirectsymoff', 'nindirectsyms',
                   'extreloff', 'nextrel', 'locreloff', 'nlocrel']
    _SIZES = [uint32_t, uint32_t, uint32_t, uint32_t,
              uint32_t, uint32_t, uint32_t, uint32_t,
              uint32_t, uint32_t, uint32_t, uint32_t,
              uint32_t, uint32_t, uint32_t, uint32_t,
              uint32_t, uint32_t, uint32_t, uint32_t]
    SIZE = 80

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)
        self.cmd = 0
        self.cmdsize = 0
        self.ilocalsym = 0
        self.nlocalsym = 0
        self.iextdefsym = 0
        self.nextdefsym = 0
        self.iundefsym = 0
        self.nundefsym = 0
        self.tocoff = 0
        self.ntoc = 0
        self.modtaboff = 0
        self.nmodtab = 0
        self.extrefsymoff = 0
        self.nextrefsyms = 0
        self.indirectsymoff = 0
        self.nindirectsyms = 0
        self.extreloff = 0
        self.nextrel = 0
        self.locreloff = 0
        self.nlocrel = 0


class dylib(Struct):
    _FIELDNAMES = ['name', 'timestamp', 'current_version', 'compatibility_version']
    _SIZES = [uint32_t, uint32_t, uint32_t, uint32_t]
    SIZE = 16

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)
        self.name = 0
        self.timestamp = 0
        self.current_version = 0
        self.compatibility_version = 0


class dylib_command(Struct):
    _FIELDNAMES = ['cmd', 'cmdsize', 'dylib']
    _SIZES = [uint32_t, uint32_t, type_bytes | 16]
    SIZE = 24

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)
        self.cmd = 0
        self.cmdsize = 0
        self.dylib = 0


class dylinker_command(Struct):
    _FIELDNAMES = ["cmd", "cmdsize", "name"]
    _SIZES = [uint32_t, uint32_t, uint32_t]
    SIZE = 12

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)
        self.cmd = 0
        self.cmdsize = 0
        self.name = 0


class sub_client_command(Struct):
    _FIELDNAMES = ['cmd', 'cmdsize', 'offset']
    _SIZES = [uint32_t, uint32_t, uint32_t]
    SIZE = 12

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)
        self.cmd = 0
        self.cmdsize = 0
        self.offset = 0


class uuid_command(Struct):
    _FIELDNAMES = ['cmd', 'cmdsize', 'uuid']
    _SIZES = [uint32_t, uint32_t, type_bytes | 16]
    SIZE = 24

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)
        self.cmd = 0
        self.cmdsize = 0
        self.uuid = 0


class build_version_command(Struct):
    _FIELDNAMES = ['cmd', 'cmdsize', 'platform', 'minos', 'sdk', 'ntools']
    _SIZES = [uint32_t, uint32_t, uint32_t, uint32_t, uint32_t, uint32_t]
    SIZE = 24

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)
        self.cmd = 0
        self.cmdsize = 0
        self.platform = 0
        self.minos = 0
        self.sdk = 0
        self.ntools = 0


class entry_point_command(Struct):
    _FIELDNAMES = ['cmd', 'cmdsize', 'entryoff', 'stacksize']
    _SIZES = [uint32_t, uint32_t, uint64_t, uint64_t]
    SIZE = 24

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)
        self.cmd = 0
        self.cmdsize = 0
        self.entryoff = 0
        self.stacksize = 0


class rpath_command(Struct):
    _FIELDNAMES = ['cmd', 'cmdsize', 'path']
    _SIZES = [uint32_t, uint32_t, uint32_t]
    SIZE = 12

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)
        self.cmd = 0
        self.cmdsize = 0
        self.path = 0


class source_version_command(Struct):
    _FIELDNAMES = ['cmd', 'cmdsize', 'version']
    _SIZES = [uint32_t, uint32_t, uint64_t]
    SIZE = sum(_SIZES)

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)
        self.cmd = 0
        self.cmdsize = 0
        self.version = 0


class linkedit_data_command(Struct):
    _FIELDNAMES = ['cmd', 'cmdsize', 'dataoff', 'datasize']
    _SIZES = [uint32_t, uint32_t, uint32_t, uint32_t]
    SIZE = sum(_SIZES)

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)
        self.cmd = 0
        self.cmdsize = 0
        self.dataoff = 0
        self.datasize = 0


class dyld_info_command(Struct):
    _FIELDNAMES = ['cmd', 'cmdsize', 'rebase_off', 'rebase_size', 'bind_off', 'bind_size',
                   'weak_bind_off', 'weak_bind_size', 'lazy_bind_off', 'lazy_bind_size',
                   'export_off', 'export_size']
    _SIZES = [uint32_t, uint32_t, uint32_t, uint32_t,
              uint32_t, uint32_t, uint32_t, uint32_t,
              uint32_t, uint32_t, uint32_t, uint32_t]
    SIZE = 48

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)
        self.cmd = 0
        self.cmdisze = 0
        self.rebase_off = 0
        self.rebase_size = 0
        self.bind_off = 0
        self.bind_size = 0
        self.weak_bind_off = 0
        self.weak_bind_size = 0
        self.lazy_bind_off = 0
        self.lazy_bind_size = 0
        self.export_off = 0
        self.export_size = 0


class symtab_entry_32(Struct):
    _FIELDNAMES = ["str_index", "type", "sect_index", "desc", "value"]
    _SIZES = [uint32_t, uint8_t, uint8_t, uint16_t, uint32_t]
    SIZE = sum(_SIZES)

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)
        self.str_index = 0
        self.type = 0
        self.sect_index = 0
        self.desc = 0
        self.value = 0


class symtab_entry(Struct):
    _FIELDNAMES = ["str_index", "type", "sect_index", "desc", "value"]
    _SIZES = [uint32_t, uint8_t, uint8_t, uint16_t, uint64_t]
    SIZE = sum(_SIZES)

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)
        self.str_index = 0
        self.type = 0
        self.sect_index = 0
        self.desc = 0
        self.value = 0


class version_min_command(Struct):
    _FIELDNAMES = ["cmd", "cmdsize", "version", "reserved"]
    _SIZES = [uint32_t, uint32_t, uint32_t, uint32_t]
    SIZE = 16

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)
        self.cmd = 0
        self.cmdsize = 0
        self.version = 0
        self.reserved = 0


class encryption_info_command(Struct):
    _FIELDNAMES = ["cmd", "cmdsize", "cryptoff", "cryptsize", "cryptid"]
    _SIZES = [uint32_t, uint32_t, uint32_t, uint32_t, uint32_t]
    SIZE = 20

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)
        self.cmd = 0
        self.cmdsize = 0
        self.cryptoff = 0
        self.cryptsize = 0
        self.cryptid = 0


class encryption_info_command_64(Struct):
    _FIELDNAMES = ["cmd", "cmdsize", "cryptoff", "cryptsize", "cryptid", "pad"]
    _SIZES = [uint32_t, uint32_t, uint32_t, uint32_t, uint32_t, uint32_t]
    SIZE = sum(_SIZES)

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)
        self.cmd = 0
        self.cmdsize = 0
        self.cryptoff = 0
        self.cryptsize = 0
        self.cryptid = 0
        self.pad = 0


class thread_command(Struct):
    _FIELDNAMES = ["cmd", "cmdsize", "flavor", "count"]
    _SIZES = [uint32_t, uint32_t, uint32_t, uint32_t]
    SIZE = sum(_SIZES)

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)
        self.cmd = 0
        self.cmdsize = 0
        self.flavor = 0
        self.count = 0



from ktool.ktool import load_image, load_objc_metadata, generate_headers, generate_text_based_stub, load_macho_file, macho_verify, reload_image, macho_combine

from ktool.objc import ObjCImage
from ktool.dyld import Dyld, LD64, Image
from ktool.macho import Slice, MachOFile, MachOFileType, Segment, Section
try:
    from ktool.headers import HeaderGenerator, Header
except ModuleNotFoundError:
    # Maybe pygments wasn't installed and we're running in some weird context
    # So let whatever works, work
    Header = None
    HeaderGenerator = None
    pass
from ktool.util import KTOOL_VERSION, ignore, log, LogLevel, Table



#!/usr/bin/env python3

#
#  ktool | MAIN SCRIPT
#  ktool
#
#  This file is the main command-line script providing utilities for using ktool.
#
#  This file is part of ktool. ktool is free software that
#  is made available under the MIT license. Consult the
#  file "LICENSE" that is distributed together with this file
#  for the exact licensing terms.
#
#  Copyright (c) kat 2021.
#

import json
import os
import os.path
import shutil
import sys
import threading
import urllib.request
from argparse import ArgumentParser
from collections import namedtuple
from enum import Enum
from typing import Union
from kimg4.img4 import IM4P

# noinspection PyProtectedMember
from pkg_resources import packaging

import ktool
from kmacho import LOAD_COMMAND

from ktool import (
    MachOFileType,
    KTOOL_VERSION,
    ignore,
    log,
    LogLevel,
    Table,
    LD64
)

from ktool.swift import *

from ktool.exceptions import *
from ktool.generator import FatMachOGenerator
from ktool.util import opts, version_output, ktool_print
from ktool.window import KToolScreen, external_hard_fault_teardown

from ktool.kcache import KernelCache, Kext, EmbeddedKext

UPDATE_AVAILABLE = False
MAIN_PARSER = None
MMAP_ENABLED = True

# noinspection PyShadowingBuiltins
print = ktool_print


def get_terminal_size():
    # We use this instead of shutil.get_terminal_size, because when output is being piped, it returns column width 80
    # We want to make sure if output is being piped (for example, to grep), that no wrapping occurs, so greps will
    # always display all relevant info on a single line. This also helps if it's being piped into a file,
    # for processing purposes among everything else.
    try:
        return os.get_terminal_size()
    except OSError:
        return shutil.get_terminal_size()


def handle_version(version: str):
    """ Used by check_for_update """
    return packaging.version.parse(version)


def check_for_update():
    endpoint = "https://pypi.org/pypi/k2l/json"
    # noinspection PyBroadException
    try:
        with urllib.request.urlopen(endpoint, timeout=1) as url:
            data = json.loads(url.read().decode(), strict=False)
        new_version = data.get('info').get('version')
        if handle_version(KTOOL_VERSION) < handle_version(new_version):
            global UPDATE_AVAILABLE
            UPDATE_AVAILABLE = True
    except Exception:
        pass


class KToolError(Enum):
    ArgumentError = 1
    FiletypeError = 2
    MalformedMachOError = 3
    ProcessingError = 4


def exit_with_error(error: KToolError, msg):
    print(f'Encountered an Error ({error.name}):\n' + f"{msg}", file=sys.stderr)
    exit(error.value)


# noinspection PyProtectedMember
def arg_dest_to_name(parser: Union[None, ArgumentParser], dest):
    """
    Convert dest (an argument destination variable name) to the actual flag used to set it (do_headers -> --headers)

    Uses internal properties from argparse, this also iterates through all subparsers

    :param parser: main argument parser to pull the original flag from
    :param dest: destination variable name
    :return:
    """
    args = {}

    for k, v in parser._option_string_actions.items():
        # parser here is an instance of ArgumentParser
        args[str(v.dest)] = k

    for parser_name, sparser in parser._subparsers._group_actions[0].choices.items():
        # if a parser has subparsers, you'll be able to get an instance of ArgumentParser representing them at
        # _subparsers._group_actions[0].choices, which is a Dict[str, ArgumentParser] containing all registered
        # subparsers.
        for k, v in sparser._option_string_actions.items():
            args[str(v.dest)] = k

    if dest not in args:
        raise AttributeError(f'{dest} destination not in any arguments.')

    return args[dest]


def require_args(args, always=None, one_of=None):
    """
    This is a quick macro to enforce argument requirements for different commands.

    If a check fails, it'll print usage for the subcommand and exit the program.

    :param args: Parsed argument object
    :param always: Arguments that *must* be passed
    :param one_of: At least one of these arguments must be passed, and must evaluate as True
    :return:
    """

    if always:
        missing = []
        for i in always:
            if not hasattr(args, i):
                missing.append(i)
            elif not getattr(args, i):
                missing.append(i)

        if len(missing) > 0:
            print(args.func.__doc__)
            if len(missing) == 1:
                error_str = f'Missing required argument {arg_dest_to_name(MAIN_PARSER, missing[0])}'
                exit_with_error(KToolError.ArgumentError, error_str)
            else:
                error_str = 'Missing required arguments: '
                error_str += ', '.join([arg_dest_to_name(MAIN_PARSER, i) for i in missing])
                exit_with_error(KToolError.ArgumentError, error_str)

    if one_of:
        found_one = False
        for i in one_of:
            if hasattr(args, i):
                if getattr(args, i):
                    found_one = True
                    break
        if not found_one:
            print(args.func.__doc__)
            missing_args = ", ".join([arg_dest_to_name(MAIN_PARSER, i) for i in one_of])
            exit_with_error(KToolError.ArgumentError, f'Missing one of {missing_args}')


def main():
    parser = ArgumentParser(description="ktool")

    parser.add_argument('--bench', dest='bench', action='store_true')
    parser.add_argument('--membench', dest='membench', action='store_true')
    parser.add_argument('-v', dest='logging_level', type=int)
    parser.add_argument('-f', dest='force_load', action='store_true')
    parser.add_argument('-V', dest='get_vers', action='store_true')
    parser.add_argument('--no-mmap', dest='no_mmap', action='store_true', help='Disable mmaped IO')
    parser.set_defaults(func=help_prompt, bench=False, membench=False, force_load=False, no_mmap=False, logging_level=1,
                        get_vers=False)

    subparsers = parser.add_subparsers(help='sub-command help')

    # open command: opens the main GUI
    parser_open = subparsers.add_parser('open', help='open ktool GUI and browse file')

    parser_open.add_argument('filename', nargs='?', default='')
    parser_open.add_argument('--hard-fail', dest='hard_fail', action='store_true')

    parser_open.set_defaults(func=_open, hard_fail=False)

    parser_json = subparsers.add_parser('json', help='Dump Image metadata as json')

    parser_json.add_argument('--with-objc', dest='with_objc', action='store_true')
    parser_json.add_argument('filename', nargs='?', default='')

    parser_json.set_defaults(func=serialize, with_objc=False)

    # insert command: optool replacement
    parser_insert = subparsers.add_parser('insert', help='Insert data into MachO Binary')

    parser_insert.add_argument('filename', nargs='?', default='')
    parser_insert.add_argument('--lc', dest='lc', help="Type of Load Command to insert")
    parser_insert.add_argument('--payload', dest='payload', help="Payload (if required) for insertion")
    parser_insert.add_argument('--out', dest='out', help="Output file destination for patches")

    parser_insert.set_defaults(func=insert, out=None, lc=None, payload=None)

    # edit command: install-name-tool replacement
    parser_edit = subparsers.add_parser('edit', help='Edit attributes of the MachO')

    parser_edit.add_argument('filename', nargs='?', default='')
    parser_edit.add_argument('--iname', dest='iname', help='Modify the Install Name of a image')
    parser_edit.add_argument('--apad', dest='apad',
                             help='Add MachO Header Padding (not yet implemented, ignore this flag please)')
    parser_edit.add_argument('--out', dest='out', help="Output file destination for patches")

    parser_edit.set_defaults(func=edit, out=None, iname=None, apad=None)

    # lipo command: you will never guess which macos cli tool the lipo command replaces
    parser_lipo = subparsers.add_parser('lipo', help='Extract/Combine slices')

    parser_lipo.add_argument('--extract', dest='extract', type=str, help='Extract a slice (--extract arm64)')
    parser_lipo.add_argument('--out', dest='out', help="Output File")
    parser_lipo.add_argument('--create', dest='combine', action='store_true',
                             help="Combine files to create a fat mach-o image")
    parser_lipo.add_argument('filename', nargs='*', default='')

    parser_lipo.set_defaults(func=lipo, out="", combine=False)

    # img4 command: because why not. wrote an img4 library for iBootLoader and it can definitely be useful elsewhere.
    parser_img4 = subparsers.add_parser('img4', help='img4/IM4P parsing utilities')

    parser_img4.add_argument('filename', nargs='?', default='')

    parser_img4.add_argument('--kbag', dest='get_kbag', action='store_true', help="Decode keybags in an im4p file")
    parser_img4.add_argument('--dec', dest='do_decrypt', action='store_true', help="Decrypt an im4p file with iv/key")
    parser_img4.add_argument('--iv', dest='aes_iv', type=str, help='IV for decryption')
    parser_img4.add_argument('--key', dest='aes_key', type=str, help='Key for decryption')
    parser_img4.add_argument('--out', dest='out', help="Output file destination for decryption")

    parser_img4.set_defaults(func=img4, get_kbag=False, do_decrypt=False, aes_iv=None, aes_key=None, out=None)

    # file command: super basic info about a file (thin/fat, and if fat, what slices are contained)
    #               replaces the relevant usage of the `file` command on macos
    parser_file = subparsers.add_parser('file', help='Print File Type (thin/fat MachO)')

    parser_file.add_argument('filename', nargs='?', default='')

    parser_file.set_defaults(func=_file)

    # info command: prints the VM map (this is honestly just for me when debugging shit)
    parser_info = subparsers.add_parser('info', help='Print Info about a MachO image')

    parser_info.add_argument('--slice', dest='slice_index', type=int,
                             help="Specify Index of Slice (in FAT MachO) to examine")
    parser_info.add_argument('--vm', dest='get_vm', action='store_true', help="Print VM Mapping for MachO image")
    parser_info.add_argument('filename', nargs='?', default='')

    parser_info.set_defaults(func=info, get_vm=False, get_lcs=False, slice_index=0)

    # dump command: for dumping headers/tbds from an image
    parser_dump = subparsers.add_parser('dump', help='Dump items (headers) from binary')

    parser_dump.add_argument('--slice', dest='slice_index', type=int,
                             help="Specify Index of Slice (in FAT MachO) to examine")
    parser_dump.add_argument('--headers', dest='do_headers', action='store_true')
    parser_dump.add_argument('--class', dest='get_class')
    parser_dump.add_argument('--use-stab-for-sel', dest='usfs', action='store_true')
    parser_dump.add_argument('--hard-fail', dest='hard_fail', action='store_true')
    parser_dump.add_argument('--sorted', dest='sort_headers', action='store_true')
    parser_dump.add_argument('--tbd', dest='do_tbd', action='store_true')
    parser_dump.add_argument('--out', dest='outdir', help="Directory to dump headers into")
    parser_dump.add_argument('filename', nargs='?', default='')

    parser_dump.set_defaults(func=dump, do_headers=False, usfs=False, sort_headers=False, do_tbd=False, slice_index=0,
                             hard_fail=False, get_class=None)

    # list command: Lists lists of things contained in lists in the image.
    parser_list = subparsers.add_parser('list', help='Print various lists')

    parser_list.add_argument('--slice', dest='slice_index', type=int,
                             help="Specify Index of Slice (in FAT MachO) to examine")
    parser_list.add_argument('--classes', dest='get_classes', action='store_true', help='Print class list')
    parser_list.add_argument('--protocols', dest='get_protos', action='store_true', help='Print Protocol list')
    parser_list.add_argument('--stypes', dest='get_swift_types', action='store_true', help='Print Swift Types')
    parser_list.add_argument('--linked', dest='get_linked', action='store_true', help='Print list of linked libraries')
    parser_list.add_argument('--cmds', dest='get_lcs', action='store_true', help="Print Load Commands")
    parser_list.add_argument('--funcs', dest='get_fstarts', action='store_true', help="Print Function Starts")
    parser_list.add_argument('filename', nargs='?', default='')

    parser_list.set_defaults(func=_list, get_lcs=False, get_classes=False, get_protos=False, get_linked=False,
                             slice_index=0, get_swift_types=False)

    # symbol command: prints various symbol tables
    parser_symbols = subparsers.add_parser('symbols', help='Print various symbols')

    parser_symbols.add_argument('--imports', dest='get_imports', action='store_true', help='Print Imports')
    parser_symbols.add_argument('--imp-acts', dest='get_actions', action='store_true', help='Print Raw Binding Imports')
    parser_symbols.add_argument('--symtab', dest='get_symtab', action='store_true', help='Print out the symtab')
    parser_symbols.add_argument('--exports', dest='get_exports', action='store_true', help='Print exports')
    parser_symbols.add_argument('--slice', dest='slice_index', type=int,
                                help="Specify Index of Slice (in FAT MachO) to examine")
    parser_symbols.add_argument('filename', nargs='?', default='')

    parser_symbols.set_defaults(func=symbols, get_imports=False, get_actions=False, get_exports=False, get_symtab=False,
                                slice_index=0)

    parser_kcache = subparsers.add_parser('kcache', help='Kernel Cache Processing')

    parser_kcache.add_argument('--info', dest='get_info', action='store_true', help='Basic KCache Info')
    parser_kcache.add_argument('--kexts', dest='get_kexts', action='store_true', help='List kexts embedded')
    parser_kcache.add_argument('--kext', dest='get_kext')
    parser_kcache.add_argument('--extract', dest='do_extract')
    parser_kcache.add_argument('filename', nargs='?', default='')

    parser_kcache.set_defaults(func=kcache, get_kext=None, get_info=False, do_extract=None, get_kexts=False)

    parser_ent = subparsers.add_parser('cs', help='Codesign processing')

    parser_ent.add_argument('--ent', dest='get_ent', action='store_true')
    parser_ent.add_argument('--slice', dest='slice_index', type=int,
                                help="Specify Index of Slice (in FAT MachO) to examine")
    parser_ent.add_argument('filename', nargs='?', default='')

    parser_ent.set_defaults(func=ent, get_ent=False, slice_index=0)

    # process the arguments the user passed us.
    # it is worth noting i set the default for `func` on each command parser to a function named without ();
    # this means when that command is used, calling args.func() will branch off to that function.
    parser.print_help = help_prompt
    args = parser.parse_args()

    global MAIN_PARSER
    MAIN_PARSER = parser

    if 'KTOOL_NO_UPDATE_CHECK' not in os.environ:
        # set this off on a separate thread before we do anything else
        # it should have time to complete by the time the rest of our code is finished
        download_thread = threading.Thread(target=check_for_update, name="UpdateChecker")
        download_thread.start()

    if args.get_vers:
        version_output()
        exit()

    if not hasattr(args, 'filename'):
        # this is our default function, bc it has no .filename attribute default set
        # They typed no command or anything, so print the program usage.
        help_prompt()
        exit()

    if not args.filename or args.filename == '':
        # if it has the .filename attribute, but its not set, then they've passed a command but with no filename
        #   so, print the usage for that command.
        print(args.func.__doc__)
        exit()

    # we minmax to -1 (show absolutely nothing) and 5 (print so much it slows down by 10x)
    log.LOG_LEVEL = LogLevel(max(min(args.logging_level, 5), -1))

    if args.force_load:
        # kind of a hack using a class attribute in ktool.util but it works.
        ignore.MALFORMED = True

    if args.no_mmap:
        global MMAP_ENABLED
        MMAP_ENABLED = False

    if args.membench:
        import tracemalloc
        tracemalloc.start(10)

        args.func(args)

        snapshot = tracemalloc.take_snapshot()
        top_stats = snapshot.statistics('lineno')

        print("[ Top 10 ]")
        for stat in top_stats[:10]:
            print(stat)

    elif args.bench:
        # This will run the program as normal, but with profiling enabled
        # It will print what are arguably the 10 routines contributing the most to slowdowns.
        import cProfile
        import pstats

        profile = cProfile.Profile()
        profile.runcall(args.func, args)
        ps = pstats.Stats(profile)
        ps.sort_stats('time', 'cumtime')  # who named this
        ps.print_stats(10)

    else:
        try:
            args.func(args)
        except UnsupportedFiletypeException:
            exit_with_error(KToolError.FiletypeError, f'{args.filename} is not a valid MachO Binary')
        except FileNotFoundError:
            exit_with_error(KToolError.ArgumentError, f'{args.filename} does not exist')
        except MalformedMachOException:
            exit_with_error(KToolError.MalformedMachOError,
                            f'Malformed MachO. Pass -f to force loading whatever possible')

    if UPDATE_AVAILABLE:
        print(f'\n\nUpdate Available ---')
        print(f'run `pip3 install --upgrade k2l` to fetch the latest update')
        print(f'set the envar KTOOL_NO_UPDATE_CHECK to disable update checks')

    exit(0)


def help_prompt():
    """Usage: ktool <global flags> [command] <flags> [filename]

Commands:

GUI (Still in active development) ---
    ktool open [filename] - Open the ktool command line GUI and browse a file

MachO Editing ---
    insert - Utils for inserting load commands into MachO Binaries
    edit - Utils for editing MachO Binaries
    lipo - Utilities for combining/separating slices in fat MachO files.

MachO Analysis ---
    dump - Tools to reconstruct certain files (headers, .tbds) from compiled MachOs
    json - Dump image metadata as json
    cs - Codesigning info
    kcache - Kernel cache specific tools
    list - Print various lists (ObjC Classes, etc.)
    symbols - Print various tables (Symbols, imports, exports)
    info - Print misc info about the target mach-o

Misc Utilities ---
    file - Print very basic info about the MachO
    img4 - IMG4 Utilities

Run `ktool [command]` for info/examples on using that command

Global Flags:
    -f - Force Load (ignores malformations in the MachO and tries to load whatever it can)
    -v [-1 through 5] - Log verbosiy. -1 completely silences logging.
        """
    print(help_prompt.__doc__)


def process_patches(image) -> 'Image':
    try:
        return ktool.reload_image(image)
    except MalformedMachOException:
        exit_with_error(KToolError.ProcessingError, "Reloading MachO after patch failed. This is an issue with "
                                                    "my patch code. Please file an issue on "
                                                    "https://github.com/kritantadev/ktool.")


def _open(args):
    try:
        log.LOG_LEVEL = LogLevel.DEBUG
        screen = KToolScreen(args.hard_fail)
        log.LOG_FUNC = screen.ktool_dbg_print_func
        log.LOG_ERR = screen.ktool_dbg_print_err_func
        screen.load_file(args.filename, MMAP_ENABLED)
    except KeyboardInterrupt:
        external_hard_fault_teardown()
        print('Hard Faulted. This was likely due to a curses error causing a freeze while rendering.')
        exit(64)
    except Exception as ex:
        external_hard_fault_teardown()
        print('Hard fault in GUI due to uncaught exception:')
        raise ex

    # should probably just always do this, just in case.
    external_hard_fault_teardown()


def serialize(args):
    """
----------
Dump image metadata as json

> ktool json [filename]

Dump image metadata including objc metadata

> ktool json --with-objc [filename]
    """
    require_args(args, one_of=['filename'])

    log.LOG_LEVEL = LogLevel(-1)

    with open(args.filename, 'rb') as fp:

        macho_file = ktool.load_macho_file(fp, use_mmaped_io=MMAP_ENABLED)

        out_dict = {
            'filetype': macho_file.type.name
        }
        slices = []

        for macho_slice in macho_file.slices:
            image = ktool.load_image(macho_slice)
            image_dict = image.serialize()
            slice_dict = {
                'offset': macho_slice.offset,
                'size': macho_slice.size,
                'type': macho_slice.type.name,
                'subtype': macho_slice.subtype.name,
                'image': image_dict
            }
            slices.append(slice_dict)

        out_dict['slices'] = slices

        if args.with_objc:
            objc_image = ktool.load_objc_metadata(image)
            out_dict['objc'] = objc_image.serialize()

        print(json.dumps(out_dict, indent=4, sort_keys=True))


def ent(args):
    """
----------
Interact with codesigning info

Dump entitlements
> ktool cs --ent [filename]
    """
    require_args(args, one_of=['get_ent'])

    if args.get_ent:
        with open(args.filename, 'rb') as fd:
            image = ktool.load_image(fd, args.slice_index, load_symtab=False, load_imports=False,
                                 use_mmaped_io=MMAP_ENABLED)
            ents = image.codesign_info.entitlements
            print(ents)


def symbols(args):
    """
----------
List symbol imports/exports

Print the list of imported symbols
> ktool symbols --imports [filename]

Print the list of exported symbols
> ktool symbols --exports [filename]

Print the symbol table
> ktool symbols --symtab [filename]
    """

    require_args(args, one_of=['get_imports', 'get_actions', 'get_exports', 'get_symtab'])

    if args.get_exports:
        with open(args.filename, 'rb') as fd:
            image = ktool.load_image(fd, args.slice_index, load_symtab=False, load_imports=False,
                                     use_mmaped_io=MMAP_ENABLED)

            table = Table()
            table.titles = ['Address', 'Symbol']

            for symbol in image.exports:
                table.rows.append([hex(symbol.address), symbol.fullname])

            print(table.fetch_all(get_terminal_size().columns))

    if args.get_symtab:
        with open(args.filename, 'rb') as fd:
            image = ktool.load_image(fd, args.slice_index, load_imports=False, load_exports=False,
                                     use_mmaped_io=MMAP_ENABLED)

            table = Table()
            table.titles = ['Address', 'Name']

            for sym in image.symbol_table.table:
                table.rows.append([hex(sym.address), sym.fullname])

            print(table.fetch_all(get_terminal_size().columns - 1))

    if args.get_imports:
        with open(args.filename, 'rb') as fd:
            image = ktool.load_image(fd, args.slice_index, load_exports=False, load_symtab=False,
                                     use_mmaped_io=MMAP_ENABLED)

            import_symbols = {}
            symbol = namedtuple('symbol', ['addr', 'name', 'image', 'from_table'])

            for addr, sym in image.import_table.items():
                try:
                    import_symbols[sym.fullname] = symbol(hex(addr), sym.fullname,
                                                          image.linked_images[int(sym.ordinal) - 1].install_name,
                                                          sym.attr)
                except IndexError:
                    import_symbols[sym.fullname] = symbol(hex(addr), sym.fullname, "ordinal: " + str(int(sym.ordinal)),
                                                          sym.attr)

            table = Table()
            table.titles = ['Addr', 'Symbol', 'Image', 'Binding']

            for _, sym in import_symbols.items():
                table.rows.append([sym.addr, sym.name, sym.image, sym.from_table])

            print(table.fetch_all(get_terminal_size().columns))

    elif args.get_actions:
        with open(args.filename, 'rb') as fd:
            image = ktool.load_image(fd, args.slice_index, use_mmaped_io=MMAP_ENABLED)

            print('\nBinding Info'.ljust(60, '-') + '\n')
            for sym in image.binding_table.symbol_table:
                try:
                    print(
                        f'{hex(sym.address).ljust(15, " ")} | '
                        f'{image.linked_images[int(sym.ordinal) - 1].install_name} | '
                        f'{sym.name.ljust(20, " ")} | {sym.dec_type}')
                except IndexError:
                    pass
            print('\nWeak Binding Info'.ljust(60, '-') + '\n')
            for sym in image.weak_binding_table.symbol_table:
                try:
                    print(
                        f'{hex(sym.address).ljust(15, " ")} | '
                        f'{image.linked_images[int(sym.ordinal) - 1].install_name} | '
                        f'{sym.name.ljust(20, " ")} | {sym.dec_type}')
                except IndexError:
                    pass
            print('\nLazy Binding Info'.ljust(60, '-') + '\n')
            for sym in image.lazy_binding_table.symbol_table:
                try:
                    print(
                        f'{hex(sym.address).ljust(15, " ")} | '
                        f'{image.linked_images[int(sym.ordinal) - 1].install_name} | '
                        f'{sym.name.ljust(20, " ")} | '
                        f'{sym.dec_type}')
                except IndexError:
                    pass


def insert(args):
    """
----------
Utils for inserting load commands into mach-o binaries

insert a LOAD_DYLIB command
> ktool insert --lc load --payload /Dylib/Install/Name/Here.dylib --out <output filename> [filename]

commands currently supported:
    load: LOAD_DYLIB
    load-weak: LOAD_WEAK_DYLIB
    lazy-load: LAZY_LOAD_DYLIB
    load-upward: LOAD_UPWARD_DYLIB
    """

    require_args(args, always=['lc'])

    lc = None
    if args.lc == "load":
        lc = LOAD_COMMAND.LOAD_DYLIB
    elif args.lc == "load-weak" or args.lc == "load_weak":
        lc = LOAD_COMMAND.LOAD_WEAK_DYLIB
    elif args.lc in ["load_lazy", "load-lazy", "lazy-load", "lazy_load"]:
        lc = LOAD_COMMAND.LAZY_LOAD_DYLIB
    elif args.lc == "load-upward" or args.lc == "load_upward":
        lc = LOAD_COMMAND.LOAD_UPWARD_DYLIB

    patched_libraries = []

    with open(args.filename, 'rb') as fp:
        macho_file = ktool.load_macho_file(fp, use_mmaped_io=MMAP_ENABLED)
        for macho_slice in macho_file.slices:

            image = ktool.load_image(macho_slice)

            last_dylib_command_index = -1
            for i, cmd in enumerate(image.macho_header.load_commands):
                if isinstance(cmd, dylib_command):
                    last_dylib_command_index = i + 1

            dylib_item = Struct.create_with_values(dylib, [0x18, 0x2, 0x010000, 0x010000])

            LD64.insert_load_cmd_with_str(image, lc, [dylib_item.raw], args.payload, last_dylib_command_index)

            log.info("Reloading MachO Slice to verify integrity")
            image = process_patches(image)
            patched_libraries.append(image)

    with open(args.out, 'wb') as fd:
        if len(patched_libraries) > 1:
            slices = [image.slice for image in patched_libraries]
            fat_generator = FatMachOGenerator(slices)
            fd.write(fat_generator.fat_head)
            for arch in fat_generator.fat_archs:
                fd.seek(arch.offset)
                fd.write(arch.slice.full_bytes_for_slice())
        else:
            fd.write(patched_libraries[0].slice.full_bytes_for_slice())


def edit(args):
    """
----------
Utils for editing MachO Binaries

Modify the install name of a image
> ktool edit --iname [Desired Install Name] --out <Output Filename> [filename]
    """

    require_args(args, one_of=['iname', 'apad'])

    patched_libraries = []

    if args.iname:

        new_iname = args.iname

        with open(args.filename, 'rb') as fp:
            macho_file = ktool.load_macho_file(fp, use_mmaped_io=MMAP_ENABLED)
            for macho_slice in macho_file.slices:
                image = ktool.load_image(macho_slice)
                id_dylib_index = -1

                for i, cmd in enumerate(image.macho_header.load_commands):
                    if cmd.cmd == 0xD:
                        id_dylib_index = i
                        break

                dylib_item = Struct.create_with_values(dylib, [0x18, 2, 0, 0])
                LD64.remove_load_command(image, id_dylib_index)

                image = process_patches(image)

                LD64.insert_load_cmd_with_str(image, LOAD_COMMAND.ID_DYLIB, [dylib_item.raw], new_iname, id_dylib_index)

                image = process_patches(image)

                patched_libraries.append(image)

        with open(args.out, 'wb') as fd:
            if len(patched_libraries) > 1:
                slices = [image.slice for image in patched_libraries]
                fat_generator = FatMachOGenerator(slices)
                fd.write(fat_generator.fat_head)
                for arch in fat_generator.fat_archs:
                    fd.seek(arch.offset)
                    fd.write(arch.slice.full_bytes_for_slice())
            else:
                fd.write(patched_libraries[0].slice.full_bytes_for_slice())


def img4(args):
    """
----------
IMG4 Utilities

Getting keybags
> ktool img4 --kbag <filename>

Decrypting an im4p
> ktool img4 --dec --iv AES_IV --key AES_KEY [--out <output-filename>] <filename>
    """

    require_args(args, one_of=['get_kbag', 'do_decrypt'])

    if args.get_kbag:
        with open(args.filename, 'rb') as fp:
            im4p = IM4P(fp.read())
            for bag in im4p.kbag.keybags:
                print(f'{bag.iv.hex()}{bag.key.hex()}')

    if args.do_decrypt:
        require_args(args, always=['aes_key', 'aes_iv'])

        out = args.out
        if not out:
            out = args.filename + '.dec'
        with open(args.filename, 'rb') as fp:
            with open(out, 'wb') as out_fp:
                im4p = IM4P(fp.read())
                out_fp.write(im4p.decrypt_data(args.aes_iv, args.aes_key))

        print(f'Attempted decrypt of data with key/iv and saved to {out}')


def lipo(args):
    """
----------
Utilities for combining/separating slices in fat MachO files.

Extract a slice from a fat binary
> ktool lipo --extract [slice_name] [filename]

Create a fat Macho Binary from multiple thin binaries
> ktool lipo --create [--out filename] [filenames]
    """

    require_args(args, one_of=['combine', 'extract'])

    if args.combine:
        output = args.out
        if output == "":
            output = args.filename[0] + '.fat'
        slices = []
        for filename in args.filename:
            # Slice() might hold a ref preventing it from being closed? but i'm just going to let it close on exit()
            fd = open(filename, 'rb')
            macho_file = ktool.load_macho_file(fd, use_mmaped_io=MMAP_ENABLED)
            if macho_file.type != MachOFileType.THIN:
                exit_with_error(KToolError.ArgumentError, "Fat mach-o passed to --create")
            slices.append(macho_file.slices[0])

        with open(output, 'wb') as fd:
            fd.write(ktool.macho_combine(slices).read())

    elif args.extract != "":
        with open(args.filename[0], 'rb') as fd:
            macho_file = ktool.load_macho_file(fd, use_mmaped_io=MMAP_ENABLED)
            output = args.out
            if output == "":
                output = args.filename[0] + '.' + args.extract.lower()
            for macho_slice in macho_file.slices:
                if macho_slice.type.name.lower() == args.extract:
                    with open(output, 'wb') as out:
                        out.write(macho_slice.full_bytes_for_slice())
                    return
            macho_slices_list = [macho_slice.type.name.lower() for macho_slice in macho_file.slices]
            exit_with_error(KToolError.ArgumentError,
                            f'Architecture {args.extract} was not found (found: {macho_slices_list})')


def _list(args):
    """
----------
Tools for printing various lists

To print the list of classes
> ktool list --classes [filename]

To print the list of protocols
> ktool list --protocols [filename]

To print a  list of linked libraries
> ktool list --linked [filename]

To print a list of Load Commands and their data
> ktool list --cmds [filename]

Print the list of function starts
> ktool list --funcs [filename]
    """

    require_args(args, one_of=['get_classes', 'get_protos', 'get_linked', 'get_lcs', 'get_swift_types', 'get_fstarts'])

    with open(args.filename, 'rb') as fd:

        if not args.get_lcs and not args.get_linked:
            image = ktool.load_image(fd, args.slice_index, use_mmaped_io=MMAP_ENABLED)
            objc_image = ktool.load_objc_metadata(image)
        else:
            image = ktool.load_image(fd, args.slice_index, False, False, False)

        if args.get_lcs:
            table = Table(dividers=True, avoid_wrapping_titles=True)
            table.titles = ['Index', 'Load Command', 'Data']
            table.size_pinned_columns = [0, 1]
            for i, lc in enumerate(image.macho_header.load_commands):
                lc_dat = str(lc)
                if LOAD_COMMAND(lc.cmd) in [LOAD_COMMAND.LOAD_DYLIB, LOAD_COMMAND.ID_DYLIB, LOAD_COMMAND.SUB_CLIENT]:
                    lc_dat += '\n"' + image.get_cstr_at(lc.off + lc.SIZE, vm=False) + '"'
                table.rows.append([str(i), LOAD_COMMAND(lc.cmd).name.ljust(15, ' '), lc_dat])
            print(table.fetch_all(get_terminal_size().columns-5))
        elif args.get_classes:
            for obj_class in objc_image.classlist:
                print(f'{obj_class.name}')
        elif args.get_swift_types:
            load_swift_types(image)
        elif args.get_protos:
            for objc_proto in objc_image.protolist:
                print(f'{objc_proto.name}')
        elif args.get_linked:
            for extlib in image.linked_images:
                print('(Weak) ' + extlib.install_name if extlib.weak else '' + extlib.install_name)
        elif args.get_fstarts:
            for addr in image.function_starts:
                print(f'Addr: {hex(addr)} -> {image.symbols[addr].fullname if addr in image.symbols else ""}')


def _file(args):
    """
----------

Print basic information about a file (e.g 'Thin MachO Binary')
> ktool file [filename]
    """
    with open(args.filename, 'rb') as fp:
        macho_file = ktool.load_macho_file(fp, use_mmaped_io=MMAP_ENABLED)

        table = Table()
        table.titles = ['Address', 'CPU Type', 'CPU Subtype']

        for macho_slice in macho_file.slices:
            table.rows.append([
                    f'{hex(macho_slice.offset)}',
                    f'{macho_slice.type.name}',
                    f'{macho_slice.subtype.name}'])

        print(table.fetch_all(get_terminal_size().columns))


def info(args):
    """
----------
Some misc info about the target mach-o

Print generic info about a MachO file
> ktool info [--slice n] [filename]

Print VM -> Slice -> Filename address mapping for a slice
of a MachO file
> ktool info [--slice n] --vm [filename]
    """
    with open(args.filename, 'rb') as fp:
        image = ktool.load_image(fp, args.slice_index, load_symtab=False, load_imports=False, load_exports=False,
                                 use_mmaped_io=MMAP_ENABLED)

        if args.get_vm:
            print(image.vm)

        else:
            message = (f'\033[32m{image.base_name} \33[37m--- \n'
                       f'\033[34mInstall Name: \33[37m{image.install_name}\n'
                       f'\033[34mFiletype: \33[37m{image.macho_header.filetype.name}\n' 
                       f'\033[34mFlags: \33[37m{", ".join([i.name for i in image.macho_header.flags])}\n'
                       f'\033[34mUUID: \33[37m{image.uuid.hex().upper()}\n'
                       f'\033[34mPlatform: \33[37m{image.platform.name}\n'
                       f'\033[34mMinimum OS: \33[37m{image.minos.x}.{image.minos.y}.{image.minos.z}\n'
                       f'\033[34mSDK Version: \33[37m{image.sdk_version.x}.{image.sdk_version.y}.{image.sdk_version.z}')

            print(message)


def dump(args):
    """
------
Tools to reconstruct certain files from compiled MachOs

Dump header for a single class
> ktool dump --class <classname> [filename]

To dump a full set of headers for a bin/framework
> ktool dump --headers --out <directory> [filename]

To dump .tbd files for a framework
> ktool dump --tbd [filename]
    """

    require_args(args, one_of=['do_headers', 'get_class', 'do_tbd'])

    if args.get_class:
        with open(args.filename, 'rb') as fp:
            image = ktool.load_image(fp, args.slice_index, use_mmaped_io=MMAP_ENABLED)

            if image.name == "":
                image.name = os.path.basename(args.filename)

            objc_image = ktool.load_objc_metadata(image)
            objc_headers = ktool.generate_headers(objc_image, sort_items=args.sort_headers)
            for header_name, header in objc_headers.items():
                if args.get_class == header_name[:-2]:
                    print(header.generate_highlighted_text())
                    break

    if args.do_headers:

        if args.hard_fail:
            ignore.OBJC_ERRORS = False

        if args.usfs:
            opts.USE_SYMTAB_INSTEAD_OF_SELECTORS = True

        with open(args.filename, 'rb') as fp:
            image = ktool.load_image(fp, args.slice_index, use_mmaped_io=MMAP_ENABLED)

            if image.name == "":
                image.name = os.path.basename(args.filename)

            objc_image = ktool.load_objc_metadata(image)

            objc_headers = ktool.generate_headers(objc_image, sort_items=args.sort_headers)

            for header_name, header in objc_headers.items():
                if not args.outdir:
                    print(f'\n\n{header_name}\n{header}')
                elif args.outdir == 'ndbg':
                    pass
                else:
                    os.makedirs(args.outdir, exist_ok=True)
                    with open(args.outdir + '/' + header_name, 'w') as out:
                        out.write(str(header))

                if args.bench:
                    pass
                    # pprint(image.bench_stats)

    elif args.do_tbd:
        with open(args.filename, 'rb') as fp:
            image = ktool.load_image(fp, args.slice_index, use_mmaped_io=MMAP_ENABLED)

            with open(image.name + '.tbd', 'w') as out_fp:
                out_fp.write(ktool.generate_text_based_stub(image, compatibility=True))


def kcache(args):
    """
------
KernelCache specific tools

List Kext IDS (And versions, and executable names if they were found)
> ktool kcache --kexts [filename]

Dump info for a specific kext
> ktool kcache --kext [Bundle ID or Executable Name] [filename]
    """
    require_args(args, one_of=['get_info', 'get_kexts', 'get_kext', 'do_extract'])

    fp = open(args.filename, 'rb')
    macho_file = ktool.load_macho_file(fp)
    kernel_cache = KernelCache(macho_file)

    if args.get_info:
        print(kernel_cache.version_str)

    elif args.get_kexts:
        for kext in kernel_cache.kexts:
            print(f'{kext.name} -> {kext.executable_name} ({kext.version})')

    elif args.get_kext:
        kext = None
        for _kext in kernel_cache.kexts:
            if args.get_kext == _kext.executable_name:
                kext = _kext
                break
        if not kext:
            for _kext in kernel_cache.kexts:
                if args.get_kext == _kext.id:
                    kext = _kext
                    break

        if isinstance(kext, Kext):
            bundle_text = f"Bundle ID: {kext.id}\nExecutable Name: {kext.executable_name}\n{kext.info_string}\nVersion: {kext.version_str}\nStart Address: {hex(kext.start_addr | 0xffff000000000000)}"
            print(bundle_text)
        else:
            print('Kext Not Found')

    elif args.do_extract:
        kext = None
        for _kext in kernel_cache.kexts:
            if args.do_extract == _kext.executable_name:
                kext = _kext
                break
        if not kext:
            for _kext in kernel_cache.kexts:
                if args.do_extract == _kext.id:
                    kext = _kext
                    break

        if isinstance(kext, EmbeddedKext):
            with open(kext.id.split('.')[-1], 'wb') as out:
                out.write(kext.image.slice.full_bytes_for_slice())
        else:
            print('Kext Not Found')


if __name__ == "__main__":
    main()



#
#  ktool | kmacho
#  base.py
#
#
#
#  This file is part of ktool. ktool is free software that
#  is made available under the MIT license. Consult the
#  file "LICENSE" that is distributed together with this file
#  for the exact licensing terms.
#
#  Copyright (c) kat 2021.
#

from abc import ABC, abstractmethod


class Constructable(ABC):
    """
    This is an attempt to define a standardized API for objects we load and may want to create.

    The idea is that all objects should be loadable and serializable in both directions, to allow patching, creation,
        and standard loading, with hopefully not too much overhead being shared between the three.

    """

    @classmethod
    @abstractmethod
    def from_image(cls, *args, **kwargs):
        """
        Base method for serializing an instance of the subclass based on raw bytes

        Implementation/Args left up to implementations, but should usually follow `from_bytes(raw: bytes)`

        :return:
        """

    @classmethod
    @abstractmethod
    def from_values(cls, *args, **kwargs):
        """
        Base method for serializing an instance of the subclass based on the required set of values to create it.

        Implementation and argument structure of this is definitely left up to subclasses.

        :return:
        """

    @abstractmethod
    def raw_bytes(self):
        """
        Built or stored raw byte representation of this item

        :return:
        """



#
#  ktool | ktool
#  headers.py
#
#  This file contains the utilities used to create ObjC Header File dumps
#
#  This file is part of ktool. ktool is free software that
#  is made available under the MIT license. Consult the
#  file "LICENSE" that is distributed together with this file
#  for the exact licensing terms.
#
#  Copyright (c) kat 2021.
#

from typing import List, Dict

from ktool.dyld import SymbolType, Image
from ktool.objc import ObjCImage, Class, Category, Protocol, Property, Method, Ivar

from ktool.util import KTOOL_VERSION

from pygments import highlight
from pygments.formatters.terminal import TerminalFormatter
from pygments.lexers.objective import ObjectiveCLexer


class HeaderUtils:

    @staticmethod
    def header_head(image: Image) -> str:
        """
        This is the prefix comments at the very top of the headers generated

        :param image: MachO Image
        :return: Newline delimited string to be placed at the top of the header.
        """
        try:
            prefix = "// Headers generated with ktool v" + KTOOL_VERSION + "\n"
            prefix += "// https://github.com/cxnder/ktool | pip3 install k2l\n"
            prefix += f'// Platform: {image.platform.name} | '
            prefix += f'Minimum OS: {image.minos.x}.{image.minos.y}.{image.minos.z} | '
            prefix += f'SDK: {image.sdk_version.x}.{image.sdk_version.y}.{image.sdk_version.z}\n\n' 
            return prefix
        except AttributeError:
            prefix = "// Headers generated with ktool v" + KTOOL_VERSION + "\n"
            prefix += "// https://github.com/cxnder/ktool | pip3 install k2l\n"
            prefix += "// Issue loading image metadata\n\n"
            return prefix


class TypeResolver:
    """
    the Type Resolver is just in charge of figuring out where imports came from.

    Initialize it with an objc image, then pass it a type name, and it'll try to figure out which
        framework that class should be imported from (utilizing the image's imports)
    """
    def __init__(self, objc_image: ObjCImage):

        self.objc_image = objc_image
        classes = []
        self.classmap = {}
        try:
            for sym in objc_image.image.imports:
                if sym.dec_type == SymbolType.CLASS:
                    self.classmap[sym.name[1:]] = sym
                    classes.append(sym)
        except AttributeError:
            pass
        self.classes = classes
        self.local_classes = objc_image.classlist
        self.local_protos = objc_image.protolist

        self._linked_cache = {'NSObject': '/System/Library/Frameworks/Foundation'}

    # noinspection PyTypeChecker
    def find_linked(self, classname: str):
        """
        given a classname, return install name of a framework if that class was imported from it.

        :param classname:
        :return:
        """

        if classname in self._linked_cache:
            return self._linked_cache[classname]

        for local in self.local_classes:
            if local.name == classname:
                self._linked_cache[classname] = ""
                return ""
        for local in self.local_protos:
            if local.name == classname[1:-1]:
                self._linked_cache[classname] = "-Protocol"
                return "-Protocol"
        if classname in self.classmap:
            try:
                name = self.objc_image.image.linked_images[int(self.classmap[classname].ordinal) - 1].install_name
                if '.dylib' in name:
                    self._linked_cache[classname] = None
                    return None
                self._linked_cache[classname] = name
                return name
            except IndexError:
                pass

        self._linked_cache[classname] = None
        return None


class HeaderGenerator:
    def __init__(self, objc_image: ObjCImage):
        self.type_resolver: TypeResolver = TypeResolver(objc_image)

        self.objc_image: ObjCImage = objc_image
        self.headers = {}

        for objc_class in objc_image.classlist:
            self.headers[objc_class.name + '.h'] = Header(self.objc_image, self.type_resolver, objc_class)
        for objc_cat in objc_image.catlist:
            if objc_cat.classname != "":
                self.headers[f'{objc_cat.classname}+{objc_cat.name}.h'] = CategoryHeader(self.objc_image, objc_cat)
        for objc_proto in objc_image.protolist:
            self.headers[objc_proto.name + '-Protocol.h'] = ProtocolHeader(self.objc_image, objc_proto)

        if self.objc_image.name == "":
            image_name = self.objc_image.image.slice.macho_file.filename
        else:
            image_name = self.objc_image.name

        self.headers[image_name + '.h'] = UmbrellaHeader(self.headers)
        self.headers[image_name + '-Structs.h'] = StructHeader(objc_image)


class StructHeader:
    def __init__(self, objc_image: ObjCImage):
        """
        Scans through structs cached in the ObjCLib's type processor and writes them to a header

        :param objc_image: image containing structs
        """
        text = ""

        for struct in objc_image.tp.structs.values():
            text += str(struct) + '\n\n'

        self.text = text

    def __str__(self):
        return self.text


class Header:
    def __init__(self, objc_image: 'ObjCImage', type_resolver, objc_class: Class):
        self.interface: Interface = Interface(objc_class)
        self.objc_image = objc_image
        self.objc_class: Class = objc_class

        self.type_resolver: TypeResolver = type_resolver

        self.forward_declaration_classes: List[str] = []
        self.forward_declaration_protocols: List[str] = []

        self.imported_classes: Dict[str, str] = {}
        self.locally_imported_classes: List[str] = []
        self.locally_imported_protocols: List[str] = []

        self._process_import_section()

        self.text = self._generate_text()
        self.highlighted_text = None

    def __str__(self):
        return self.text

    def generate_highlighted_text(self):
        if self.highlighted_text:
            return self.highlighted_text

        formatter = TerminalFormatter()
        self.highlighted_text = highlight(self.text, ObjectiveCLexer(), formatter)

        return self.highlighted_text

    def _generate_text(self) -> str:
        """
        Generates the header text based on the processed and configured properties

        :return: the header text
        """
        text = [HeaderUtils.header_head(self.objc_image.image),
                "#ifndef " + self.objc_class.name.upper() + "_H",
                "#define " + self.objc_class.name.upper() + "_H",
                ""]

        for i in self.objc_class.load_errors:
            text.append(f'// {i}')

        if len(self.objc_class.load_errors) > 0:
            text.append('')

        if len(self.forward_declaration_classes) > 0:
            text.append("@class " + ", ".join(self.forward_declaration_classes) + ";")
        if len(self.forward_declaration_protocols) > 0:
            text.append("@protocol " + ", ".join(self.forward_declaration_protocols) + ";")

        text.append("")

        imported_classes = {}

        for objc_class, install_name in self.imported_classes.items():
            if '/Frameworks/' in install_name:
                nam = install_name.split("/")[-1]
                if nam not in imported_classes:
                    imported_classes[nam] = nam
            else:
                imported_classes[objc_class] = install_name

        for objc_class, install_name in imported_classes.items():
            text.append(f'#import <{install_name.split("/")[-1]}/{objc_class}.h>')

        text.append("")

        for objc_class in self.locally_imported_classes:
            text.append(f'#import "{objc_class}.h"')

        for objc_protocol in self.locally_imported_protocols:
            text.append(f'#import "{objc_protocol}-Protocol.h"')

        text.append("")

        text.append(str(self.interface))
        text.append("")
        text.append("")

        text.append("#endif")

        return "\n".join(text)

    def _process_import_section(self):
        if self.interface.objc_class.superclass != "":
            type_name = self.interface.objc_class.superclass.split('_')[-1]
            resolved_type = self.type_resolver.find_linked(type_name)
            if resolved_type is None:
                if type_name != "id":
                    if type_name.startswith('<'):
                        if type_name[1:-1] not in self.forward_declaration_protocols:
                            self.forward_declaration_protocols.append(type_name[1:-1])
                    elif type_name.startswith('NSObject<'):
                        if type_name[9:-1] not in self.forward_declaration_protocols:
                            self.forward_declaration_protocols.append(type_name[9:-1])
                    else:
                        if type_name not in self.forward_declaration_classes:
                            self.forward_declaration_classes.append(type_name)
            elif resolved_type == "":
                if type_name not in self.locally_imported_classes:
                    self.locally_imported_classes.append(type_name)
            else:
                if type_name not in self.imported_classes:
                    self.imported_classes[type_name] = resolved_type
        for protocol in self.interface.objc_class.protocols:
            type_name = f'<{protocol.name}>'
            resolved_type = self.type_resolver.find_linked(type_name)
            if resolved_type == "-Protocol":
                self.locally_imported_protocols.append(protocol.name)
            else:
                self.forward_declaration_protocols.append(protocol.name)
        for ivar in self.interface.ivars:
            if ivar.is_id:
                type_name = ivar.type
                resolved_type = self.type_resolver.find_linked(type_name)
                if resolved_type is None:
                    if type_name != "id":
                        if type_name.startswith('<'):
                            if type_name[1:-1] not in self.forward_declaration_protocols:
                                self.forward_declaration_protocols.append(type_name[1:-1])
                        elif type_name.startswith('NSObject<'):

                            if type_name[9:-1] not in self.forward_declaration_protocols:
                                self.forward_declaration_protocols.append(type_name[9:-1])
                        else:
                            if type_name not in self.forward_declaration_classes:
                                self.forward_declaration_classes.append(type_name)
                elif resolved_type == "":
                    if type_name not in self.locally_imported_classes:
                        self.locally_imported_classes.append(type_name)
                elif resolved_type == "-Protocol":
                    if type_name not in self.locally_imported_protocols:
                        self.locally_imported_protocols.append(type_name[1:-1])
                else:
                    if type_name not in self.imported_classes:
                        self.imported_classes[type_name] = resolved_type
        for objc_property in self.interface.properties:
            if objc_property.is_id:
                type_name = objc_property.type
                resolved_type = self.type_resolver.find_linked(type_name)
                if resolved_type is None:
                    if type_name != "id":
                        if type_name.startswith('<'):
                            if type_name[1:-1] not in self.forward_declaration_protocols:
                                self.forward_declaration_protocols.append(type_name[1:-1])
                        elif type_name.startswith('NSObject<'):

                            if type_name[9:-1] not in self.forward_declaration_protocols:
                                self.forward_declaration_protocols.append(type_name[9:-1])
                        else:
                            if type_name not in self.forward_declaration_classes:
                                self.forward_declaration_classes.append(type_name)
                elif resolved_type == "":
                    if type_name not in self.locally_imported_classes:
                        self.locally_imported_classes.append(type_name)
                elif resolved_type == "-Protocol":
                    if type_name not in self.locally_imported_protocols:
                        self.locally_imported_protocols.append(type_name[1:-1])
                else:
                    if type_name not in self.imported_classes:
                        self.imported_classes[type_name] = resolved_type


class CategoryHeader:
    def __init__(self, objc_image, objc_category: Category):
        self.objc_image = objc_image
        self.category = objc_category

        self.properties = objc_category.properties
        self.methods = objc_category.methods
        self.protocols = objc_category.protocols

        self.interface = CategoryInterface(objc_category)

        self.text = self._generate_text()

    def __str__(self):
        return self.text

    def _generate_text(self):
        """
        Generate Category text

        :return: category text
        """
        text = [HeaderUtils.header_head(self.objc_image.image),
                "",
                str(self.interface),
                "",
                ""]

        return "\n".join(text)


class ProtocolHeader:
    def __init__(self, objc_image, objc_protocol: Protocol):
        self.objc_image = objc_image
        self.protocol: Protocol = objc_protocol

        self.interface = ProtocolInterface(objc_protocol)

        self.text = self._generate_text()

    def __str__(self):
        return self.text

    def _generate_text(self):
        """
        Generate Protocol Header text

        :return:
        """
        text = [HeaderUtils.header_head(self.objc_image.image),
                "",
                str(self.interface),
                "",
                ""]

        return "\n".join(text)


class Interface:
    """
    The Interface class represents the "main body" of the header (not including the import section)

    """
    def __init__(self, objc_class: Class):
        self.objc_class = objc_class

        self.properties: List[Property] = []
        self.methods: List[Method] = []
        self.ivars: List[Ivar] = []
        self.structs = []

        # just store these so we know not to display them
        self.getters: List[str] = []
        self.setters: List[str] = []
        self._process_properties()
        self._process_methods()
        self._process_ivars()

    def __str__(self):
        head = "@interface " + self.objc_class.name + ' : '

        # Decode Superclass Name
        superclass = "NSObject"
        if self.objc_class.superclass != "":  # _OBJC_CLASS_$_UIApplication
            superclass = self.objc_class.superclass.split('_')[-1]

        head += superclass

        # Protocol Implementing Declaration
        if len(self.objc_class.protocols) > 0:
            head += " <"
            for protocol in self.objc_class.protocols:
                head += str(protocol) + ', '
            head = head[:-2]
            head += '>\n\n'

        # Ivar Declaration
        ivars = ""
        if len(self.ivars) > 0:
            ivars = " {\n"
            for ivar in self.ivars:
                ivars += '    ' + str(ivar) + ';\n'
            ivars += '}\n'

        props = "\n\n"
        for prop in self.properties:
            props += str(prop) + ';'
            if prop.ivarname != "":
                props += ' // ivar: ' + prop.ivarname + '\n'
            else:
                props += '\n'

        meths = "\n\n"
        for i in self.methods:
            if '.cxx_' not in str(i):
                meths += str(i) + ';\n'

        foot = "\n\n@end"
        return head + ivars + props + meths + foot

    def _process_properties(self):
        for objc_property in self.objc_class.properties:
            if not hasattr(objc_property, 'type'):
                continue
            if objc_property.type.lower() == 'bool':
                getter_name = 'is' + objc_property.name[0].upper() + objc_property.name[1:]
                self.getters.append(getter_name)
            else:
                self.getters.append(objc_property.name)
            if 'readonly' not in objc_property.attributes:
                setter_name = 'set' + objc_property.name[0].upper() + objc_property.name[1:]
                self.setters.append(setter_name)
            self.properties.append(objc_property)

    def _process_ivars(self):
        for ivar in self.objc_class.ivars:
            bad = False
            for prop in self.properties:
                if ivar.name == prop.ivarname:
                    bad = True
                    break
            if bad:
                continue
            self.ivars.append(ivar)

    def _process_methods(self):
        for method in self.objc_class.methods:
            bad = False
            for name in self.getters:
                if name in method.sel and ':' not in method.sel:
                    bad = True
                    break
            for name in self.setters:
                if name in method.sel and 'set' in method.sel:
                    bad = True
                    break
            if bad:
                continue
            self.methods.append(method)


class StructDef:
    def __init__(self, struct_definition):
        self.struct_definition = struct_definition


class CategoryInterface:
    def __init__(self, objc_category: Category):
        self.category = objc_category

        self.properties = self.category.properties
        self.methods = self.category.methods
        self.protocols = self.category.protocols

        self.text = self._generate_text()

    def __str__(self):
        return self.text

    def _generate_text(self):

        head = "@interface "

        head += self.category.classname + " (" + self.category.name + ")"

        # Protocol Implementing Declaration
        if len(self.category.protocols) > 0:
            head += " <"
            for protocol in self.category.protocols:
                head += str(protocol) + ', '
            head = head[:-2]
            head += '>\n'

        # Ivar Declaration

        props = "\n\n"
        for prop in self.properties:
            props += str(prop) + ';'
            if prop.ivarname != "":
                props += ' // ivar: ' + prop.ivarname + '\n'
            else:
                props += '\n'

        meths = "\n\n"
        for i in self.methods:
            meths += str(i) + ';\n'

        foot = "@end\n"

        return head + props + meths + foot


class ProtocolInterface:
    def __init__(self, protocol: Protocol):
        self.protocol: Protocol = protocol

        self.text = self._generate_text()

    def __str__(self):
        return self.text

    def _generate_text(self):
        text = ["@protocol " + self.protocol.name, ""]

        for prop in self.protocol.properties:
            pro = ""
            pro += str(prop) + ';'
            if hasattr(prop, 'ivarname'):
                if prop.ivarname != "":
                    pro += ' // ivar: ' + prop.ivarname + ''
                else:
                    pro += ''
            text.append(pro)

        text.append("")

        for meth in self.protocol.methods:
            text.append(str(meth) + ';')

        text.append("")

        if len(self.protocol.opt_methods) > 0:
            text.append("@optional")
            for meth in self.protocol.opt_methods:
                text.append(str(meth) + ';')

        text.append("@end")

        return "\n".join(text)


class UmbrellaHeader:
    def __init__(self, header_list: dict):
        """
        Generates a header that solely imports other headers

        :param header_list: Dict of headers to be imported
        """
        self.text = "\n\n"
        for header in header_list.keys():
            self.text += "#include \"" + header + "\"\n"

    def __str__(self):
        return self.text



#
#  ktool | kmacho
#  codesign.py
#
#
#
#  This file is part of ktool. ktool is free software that
#  is made available under the MIT license. Consult the
#  file "LICENSE" that is distributed together with this file
#  for the exact licensing terms.
#
#  Copyright (c) kat 2022.
#

from kmacho.structs import Struct, uint32_t

CSMAGIC_REQUIREMENT = 0xfade0c00
CSMAGIC_REQUIREMENTS           = 0xfade0c01
CSMAGIC_CODEDIRECTORY          = 0xfade0c02
CSMAGIC_EMBEDDED_SIGNATURE     = 0xfade0cc0
CSMAGIC_EMBEDDED_SIGNATURE_OLD = 0xfade0b02
CSMAGIC_EMBEDDED_ENTITLEMENTS  = 0xfade7171
CSMAGIC_EMBEDDED_DERFORMAT     = 0xfade7172
CSMAGIC_DETACHED_SIGNATURE     = 0xfade0cc1
CSMAGIC_BLOBWRAPPER            = 0xfade0b01

CSSLOT_CODEDIRECTORY = 0x00000
CSSLOT_INFOSLOT      = 0x00001
CSSLOT_REQUIREMENTS  = 0x00002
CSSLOT_RESOURCEDIR   = 0x00003
CSSLOT_APPLICATION   = 0x00004
CSSLOT_ENTITLEMENTS  = 0x00005
CSSLOT_REPSPECIFIC   = 0x00006
CSSLOT_DERFORMAT     = 0x00007
CSSLOT_ALTERNATE     = 0x01000


class BlobIndex(Struct):
    _FIELDNAMES = ["type", "offset"]
    _SIZES = [uint32_t, uint32_t]
    SIZE = 8

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)
        self.type = 0
        self.offset = 0


class Blob(Struct):
    _FIELDNAMES = ["magic", "length"]
    _SIZES = [uint32_t, uint32_t]
    SIZE = 8

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)
        self.magic = 0
        self.length = 0


class SuperBlob(Struct):
    _FIELDNAMES = ["blob", "count"]
    _SIZES = [Blob, uint32_t]
    SIZE = 12

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)
        self.blob = 0
        self.count = 0



#
#  ktool | kswift
#  __init__.py
#
#
#
#  This file is part of ktool. ktool is free software that
#  is made available under the MIT license. Consult the
#  file "LICENSE" that is distributed together with this file
#  for the exact licensing terms.
#
#  Copyright (c) kat 2022.
#



#
#  ktool | ktool
#  ktool.py
#
#  Outward facing API
#
#  Some of these functions are only one line long, but the point is to standardize an outward facing API that allows
#  me to refactor and change things internally without breaking others' scripts.
#
#  This file is part of ktool. ktool is free software that
#  is made available under the MIT license. Consult the
#  file "LICENSE" that is distributed together with this file
#  for the exact licensing terms.
#
#  Copyright (c) kat 2021.
#

from typing import Dict, Union, BinaryIO, List
from io import BytesIO

from ktool.dyld import Dyld, Image
from ktool.generator import TBDGenerator, FatMachOGenerator
try:
    from ktool.headers import HeaderGenerator, Header
except ModuleNotFoundError:
    # Maybe pygments wasn't installed and we're running in some weird context
    # So let whatever works, work
    Header = None
    HeaderGenerator = None
    pass
from ktool.macho import Slice, MachOFile, SlicedBackingFile
from ktool.objc import ObjCImage
from ktool.util import TapiYAMLWriter, ignore, log


def load_macho_file(fp: Union[SlicedBackingFile, BinaryIO, BytesIO], use_mmaped_io=True) -> MachOFile:
    """
    This function takes a bare file and loads it as a MachOFile.

    File should be opened with 'rb'

    :param fp: BinaryIO object
    :param use_mmaped_io: Should the MachOFile be loaded with a mmaped-io-backend? Leaving this enabled massively
                            improves load time and IO performance, only disable if your system doesn't support it
    :return:
    """
    if isinstance(fp, BytesIO):
        use_mmaped_io = False
    elif isinstance(fp, SlicedBackingFile):
        use_mmaped_io = False
        new_fp = BytesIO()
        new_fp.write(bytes(fp.file))
        new_fp.seek(0)
        fp = new_fp

    return MachOFile(fp, use_mmaped_io=use_mmaped_io)


def reload_image(image: Image) -> Image:
    """
    Reload an image (properly updates internal representations after patches)

    :param image:
    :return:
    """
    # This is going to be horribly slow. Dyld class needs refactored to have a better way to do this or ideally just
    #   not fuck things up and require a reload every time we make a patch.
    return load_image(image.slice)


def load_image(fp: Union[BinaryIO, MachOFile, Slice, BytesIO, SlicedBackingFile], slice_index=0, load_symtab=True, load_imports=True,
               load_exports=True, use_mmaped_io=True) -> Image:
    """
    Take a bare file, MachOFile, BytesIO, SlicedBackingFile, or Slice, and load MachO/dyld metadata about that item

    :param fp: a bare file, MachOFile, or Slice to load.
    :param slice_index: If a Slice is not being passed, and a file or MachOFile is a Fat MachO, which slice should be loaded?
    :param use_mmaped_io: If a bare file is being passed, load it with mmaped IO?
    :param load_symtab: Load the symbol table if one exists. This can be disabled for targeted loads, for speed.
    :param load_imports: Load imports if they exist. This can be disabled for targeted loads, for speed.
    :param load_exports: Load exports if they exist. This can be disabled for targeted loads, for speed.
    :return: Returns a loaded Image object
    :rtype: Image
    """
    if isinstance(fp, MachOFile):
        macho_file = fp
        macho_slice: Slice = macho_file.slices[slice_index]
    elif isinstance(fp, Slice):
        macho_slice = fp
    elif isinstance(fp, BytesIO):
        macho_file = load_macho_file(fp, use_mmaped_io=False)
        macho_slice: Slice = macho_file.slices[slice_index]
    elif isinstance(fp, SlicedBackingFile):
        macho_file = load_macho_file(fp, use_mmaped_io=False)
        macho_slice: Slice = macho_file.slices[slice_index]
    else:
        macho_file = load_macho_file(fp, use_mmaped_io=use_mmaped_io)
        macho_slice: Slice = macho_file.slices[slice_index]

    return Dyld.load(macho_slice, load_symtab=load_symtab, load_imports=load_imports, load_exports=load_exports)


def macho_verify(fp: Union[BinaryIO, MachOFile, Slice, Image]) -> None:
    """
    This function takes a variety of MachO-based objects, and loads them with malformation exceptions fully enabled.

    This can be used to verify patch code did not damage or improperly modify a MachO.

    :param fp: One of: BinaryIO, MachOFile, Slice, or Image, to load and verify
    :return:
    :raises: MalformedMachOException
    """
    should_ignore = ignore.MALFORMED

    log.info("Verifying MachO Integrity")
    ignore.MALFORMED = False

    if isinstance(fp, Image):
        load_image(fp.slice)

    elif isinstance(fp, MachOFile) or isinstance(fp, BinaryIO):
        if isinstance(fp, MachOFile):
            slices = fp.slices
        else:
            slices = load_macho_file(fp)

        for macho_slice in slices:
            load_image(macho_slice)

    else:
        load_image(fp)

    ignore.MALFORMED = should_ignore


def load_objc_metadata(image: Image) -> ObjCImage:
    return ObjCImage.from_image(image)


def generate_headers(objc_image: ObjCImage, sort_items=False) -> Dict[str, Header]:
    out = {}

    if sort_items:
        for objc_class in objc_image.classlist:
            objc_class.methods.sort(key=lambda h: h.signature)
            objc_class.properties.sort(key=lambda h: h.name)

        for objc_proto in objc_image.protolist:
            objc_proto.methods.sort(key=lambda h: h.signature)
            objc_proto.opt_methods.sort(key=lambda h: h.signature)

    for header_name, header in HeaderGenerator(objc_image).headers.items():
        out[header_name] = header

    return out


def generate_text_based_stub(image: Image, compatibility=True) -> str:
    generator = TBDGenerator(image, compatibility)
    return TapiYAMLWriter.write_out(generator.dict)


def macho_combine(slices: List[Slice]) -> BytesIO:
    fat_generator = FatMachOGenerator(slices)

    fat_file = BytesIO()
    fat_file.write(fat_generator.fat_head)

    for arch in fat_generator.fat_archs:
        fat_file.seek(arch.offset)
        fat_file.write(arch.slice.full_bytes_for_slice())

    fat_file.seek(0)
    return fat_file



#
#  ktool | ktool
#  swift.py
#
#  Swift type processing
#
#  This file is part of ktool. ktool is free software that
#  is made available under the MIT license. Consult the
#  file "LICENSE" that is distributed together with this file
#  for the exact licensing terms.
#
#  Copyright (c) kat 2021.
#

import ktool
from kmacho.base import Constructable
from kswift.structs import *
from kswift.demangle import demangle
from ktool import ObjCImage

from ktool.dyld import Image
from ktool.macho import Section
from ktool.objc import Class
from ktool.util import usi32_to_si32


class Field:
    def __init__(self, flags, type_name, name):
        self.flags = flags
        self.type_name = type_name
        self.name = name

    def __str__(self):
        return f'{self.name} : {self.type_name} ({hex(self.flags)})'


class _FieldDescriptor(Constructable):

    @classmethod
    def from_image(cls, image, location):

        fields = []

        fd = image.load_struct(location, FieldDescriptor, vm=True)

        for i in range(fd.numFields):
            ea = location + (i * 0xc)
            record = image.load_struct(ea, FieldRecord, vm=True)

            flags = record.Flags
            type_name_loc = ea + 4 + record.Type
            name_loc = ea + 8 + record.Name
            try:
                name = image.get_cstr_at(name_loc, vm=True)
            except ValueError:
                name = ""
            try:
                type_name = image.get_cstr_at(type_name_loc, vm=True)
            except ValueError:
                type_name = ""

            fields.append(Field(flags, type_name, name))

        return cls(fields, fd)

    @classmethod
    def from_values(cls, *args, **kwargs):
        pass

    def raw_bytes(self):
        pass

    def __init__(self, fields, desc):
        self.fields = fields
        self.desc = desc


class SwiftClass(Constructable):

    @classmethod
    def from_image(cls, image: Image, objc_image: ObjCImage, type_location):
        class_descriptor = image.load_struct(type_location, NominalClassDescriptor, vm=True)
        name = image.get_cstr_at(type_location + 8 + class_descriptor.Name, vm=True)
        fd_loc = class_descriptor.Fields + type_location + 16
        field_descriptor = _FieldDescriptor.from_image(image, fd_loc)
        ivars = []

        for objc_class in objc_image.classlist:
            mangled_name = objc_class.name
            project, classname = demangle(mangled_name)
            if classname == name:
                objc_backing_class: Class = objc_class
                ivars = objc_backing_class.ivars
                name = f'{project}.{name}'

        return cls(name, field_descriptor.fields, class_descriptor, field_descriptor, ivars)

    @classmethod
    def from_values(cls, *args, **kwargs):
        pass

    def raw_bytes(self):
        pass

    def __init__(self, name, fields, class_descriptor=None, field_descriptor=None, ivars=None):
        self.name = name
        self.fields = fields
        self.class_desc = class_descriptor
        self.field_desc = field_descriptor
        self.ivars = ivars


class SwiftType(Constructable):

    @classmethod
    def from_image(cls, image: Image, objc_image, type_location):

        typedesc = image.load_struct(type_location, NominalTypeDescriptor, vm=True)
        name = image.get_cstr_at(type_location + 8 + typedesc.mangledName, vm=True)
        kind = ContextDescriptorKind(image.get_int_at(typedesc.off, 1, vm=False) & 0x1f)

        if kind == ContextDescriptorKind.Class:
            return SwiftClass.from_image(image, objc_image, type_location)

        fd_loc = typedesc.Fields + type_location + 16
        field_descriptor = _FieldDescriptor.from_image(image, fd_loc)

        return cls(name, kind, typedesc, field_descriptor)

    @classmethod
    def from_values(cls, *args, **kwargs):
        pass

    def raw_bytes(self):
        pass

    def __init__(self, name, kind, typedesc=None, field_desc=None):
        self.name = name
        self.kind = kind
        self.typedesc = typedesc
        self.field_desc = field_desc


def load_swift_types(image: Image):

    objc_image = ktool.load_objc_metadata(image)

    swift_type_seg_start_sect: Section = image.segments['__TEXT'].sections['__swift5_types']
    sect_start = swift_type_seg_start_sect.vm_address
    sect_size = swift_type_seg_start_sect.size

    types = []

    for ea in range(sect_start, sect_start + sect_size, 4):
        typeref = usi32_to_si32(image.get_int_at(ea, 4, vm=True))
        type_loc = ea + typeref
        types.append(SwiftType.from_image(image, objc_image, type_loc))

    return types



#
#  ktool | kswift
#  structs.py
#
#
#
#  This file is part of ktool. ktool is free software that
#  is made available under the MIT license. Consult the
#  file "LICENSE" that is distributed together with this file
#  for the exact licensing terms.
#
#  Copyright (c) kat 2022.
#


"""
struct _NominalTypeDescriptor {
    var property1: Int32
    var property2: Int32
    var mangledName: Int32
    var property4: Int32
    var numberOfFields: Int32
    var fieldOffsetVector: Int32
}"""
import enum

from kmacho.structs import *

"""
struct _FieldDescriptor {
    var mangledTypeNameOffset: Int32
    var superClassOffset: Int32
    var fieldDescriptorKind: FieldDescriptorKind
    var fieldRecordSize: Int16
    var numFields: Int32
}
"""


class FieldDescriptor(Struct):
    _FIELDNAMES = ['mangledTypeNameOffset', 'superClassOffset', 'fieldDescriptorKind', 'fieldRecordSize', 'numFields']
    _SIZES = [int32_t, int32_t, uint16_t, int16_t, int32_t]
    SIZE = 16

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order, no_patch=True)


"""
    var fieldRecordFlags: Int32
    var mangledTypeNameOffset: Int32
    var fieldNameOffset: Int32"""


class FieldRecord(Struct):
    _FIELDNAMES = ['Flags', 'Type', 'Name']
    _SIZES = [uint32_t, int32_t, int32_t]
    SIZE = 0xc

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order, no_patch=True)


class NominalTypeDescriptor(Struct):
    _FIELDNAMES = ['property1', 'property2', 'mangledName', 'property4', 'Fields', 'fieldOffsetVector']
    _SIZES = [uint32_t, int32_t, int32_t, int32_t, int32_t, int32_t]
    SIZE = 24

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order, no_patch=True)


class NominalClassDescriptor(Struct):
    _FIELDNAMES = ['Flags', 'Parent', 'Name', 'AccessFunctionPtr', 'Fields', 'SuperclassType',
                   'MetadataNegativeSizeInWords', 'MetadataPositiveSizeInWords', 'NumImmediateMembers',
                   'NumFields', 'FieldOffsetVectorOffset']
    _SIZES = [uint32_t, int32_t, int32_t, int32_t, int32_t, int32_t, uint32_t, uint32_t, uint32_t, uint32_t, uint32_t]
    SIZE = 44

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order, no_patch=True)


class ClassMethodListTable(Struct):
    _FIELDNAMES = ['VTableOffset', 'VTableSize']
    _SIZES = [4, 4]
    SIZE = 8

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order, no_patch=True)


class TargetMethodDescriptor(Struct):
    _FIELDNAMES = ['Flags', 'Impl']
    _SIZES = [uint32_t, int32_t]
    SIZE = 8

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order, no_patch=True)


class ContextDescriptorKind(enum.Enum):
    Module = 0
    Extension = 1
    Anonymous = 2
    SwiftProtocol = 3
    OpaqueType = 4
    Class = 16
    Struct = 17
    Enum = 18
    Type_Last = 31



#
#  ktool | kmacho
#  fixups.py
#
#
#
#  This file is part of ktool. ktool is free software that
#  is made available under the MIT license. Consult the
#  file "LICENSE" that is distributed together with this file
#  for the exact licensing terms.
#
#  Copyright (c) kat 2021.
#

from enum import Enum

from kmacho.structs import *


class dyld_chained_fixups_header(Struct):
    _FIELDNAMES = ['fixups_version', 'starts_offset', 'imports_offset', 'symbols_offset', 'imports_count',
                   'imports_format', 'symbols_format']
    _SIZES = [4, 4, 4, 4, 4, 4, 4]
    SIZE = sum(_SIZES)

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)


class dyld_chained_starts_in_image(Struct):
    _FIELDNAMES = ['seg_count', 'seg_info_offset']
    _SIZES = [4, 4]
    SIZE = sum(_SIZES)

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)


class dyld_chained_starts_in_segment(Struct):
    _FIELDNAMES = ['size', 'page_size', 'pointer_format', 'segment_offset', 'max_valid_pointer', 'page_count',
                   'page_starts']
    _SIZES = [4, 1, 1, 4, 2, 1, 1]
    SIZE = sum(_SIZES)

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)


DYLD_CHAINED_PTR_START_NONE = 0xFFFF
DYLD_CHAINED_PTR_START_MULTI = 0x8000
DYLD_CHAINED_PTR_START_LAST = DYLD_CHAINED_PTR_START_MULTI


class dyld_chained_start_offsets(Struct):
    _FIELDNAMES = ['pointer_format', 'starts_count', 'chain_starts']
    _SIZES = [2, 2, 2]
    SIZE = sum(_SIZES)

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)


"""
enum {
    DYLD_CHAINED_PTR_ARM64E                 =  1,    // stride 8, unauth target is vmaddr
    DYLD_CHAINED_PTR_64                     =  2,    // target is vmaddr
    DYLD_CHAINED_PTR_32                     =  3,
    DYLD_CHAINED_PTR_32_CACHE               =  4,
    DYLD_CHAINED_PTR_32_FIRMWARE            =  5,
    DYLD_CHAINED_PTR_64_OFFSET              =  6,    // target is vm offset
    DYLD_CHAINED_PTR_ARM64E_OFFSET          =  7,    // old name
    DYLD_CHAINED_PTR_ARM64E_KERNEL          =  7,    // stride 4, unauth target is vm offset
    DYLD_CHAINED_PTR_64_KERNEL_CACHE        =  8,
    DYLD_CHAINED_PTR_ARM64E_USERLAND        =  9,    // stride 8, unauth target is vm offset
    DYLD_CHAINED_PTR_ARM64E_FIRMWARE        = 10,    // stride 4, unauth target is vmaddr
    DYLD_CHAINED_PTR_X86_64_KERNEL_CACHE    = 11,    // stride 1, x86_64 kernel caches
    there is apparently a 12. it is not in xnu source. i found it in an arm64e userland (Console.app M1) bin, so we're assuming its
                                                        like, 1 I guess. 
};"""


class dyld_chained_ptr_format(Enum):
    DYLD_CHAINED_PTR_ARM64E = 1
    DYLD_CHAINED_PTR_64 = 2
    DYLD_CHAINED_PTR_32 = 3
    DYLD_CHAINED_PTR_32_CACHE = 4
    DYLD_CHAINED_PTR_32_FIRMWARE = 5
    DYLD_CHAINED_PTR_64_OFFSET = 6
    DYLD_CHAINED_PTR_ARM64E_OFFSET = 7
    DYLD_CHAINED_PTR_ARM64E_KERNEL = 7
    DYLD_CHAINED_PTR_64_KERNEL_CACHE = 8
    DYLD_CHAINED_PTR_ARM64E_USERLAND = 9
    DYLD_CHAINED_PTR_ARM64E_FIRMWARE = 10
    DYLD_CHAINED_PTR_x86_64_KERNEL_CACHE = 11
    DYLD_CHAINED_PTR_NOIDEA = 12


class dyld_chained_import_format(Enum):
    DYLD_CHAINED_IMPORT = 1
    DYLD_CHAINED_IMPORT_ADDEND = 2
    DYLD_CHAINED_IMPORT_ADDEND64 = 3


class dyld_chained_ptr(Struct):
    _FIELDNAMES = ['ptr']
    _SIZES = [8]
    SIZE = 8

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)


def ptr_arm64e_type(ptr: 'dyld_chained_ptr_arm64e'):
    if ptr.auth:
        if ptr.bind:
            return dyld_chained_ptr_arm64e_auth_bind
        else:
            return dyld_chained_ptr_arm64e_auth_rebase
    else:
        if ptr.bind:
            return dyld_chained_ptr_arm64e_bind
        else:
            return dyld_chained_ptr_arm64e_rebase


def ptr_arm64_type(ptr: 'dyld_chained_ptr'):
    if ptr.bind:
        return dyld_chained_ptr_64_bind
    return dyld_chained_ptr_64_rebase


DYLD_CHAINED_PTR_FMATS = {
    dyld_chained_ptr_format.DYLD_CHAINED_PTR_ARM64E: ptr_arm64e_type,
    dyld_chained_ptr_format.DYLD_CHAINED_PTR_64: ptr_arm64_type,
    dyld_chained_ptr_format.DYLD_CHAINED_PTR_NOIDEA: ptr_arm64e_type
}


class dyld_chained_ptr_arm64e(BitFieldStruct):
    _FIELDNAMES = ['value']
    _BITFIELDS = ['reserved', 'bind', 'auth']
    _SIZES = [8]
    _BF_SIZES = [62, 1, 1]
    SIZE = 8

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)
        self.bind = 0
        self.auth = 0


class dyld_chained_ptr_arm64e_rebase(BitFieldStruct):
    _FIELDNAMES = ['value']
    _BITFIELDS = ['target', 'high8', 'next', 'bind', 'auth']
    _SIZES = [8]
    _BF_SIZES = [43, 8, 11, 1, 1]
    SIZE = 8

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)


class dyld_chained_ptr_arm64e_bind(BitFieldStruct):
    _FIELDNAMES = ['value']
    _BITFIELDS = ['ordinal', 'zero', 'addend', 'next', 'bind', 'auth']
    _SIZES = [uint64_t]
    _BF_SIZES = [16, 16, 19, 11, 1, 1]
    SIZE = uint64_t

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)


class dyld_chained_ptr_arm64e_auth_rebase(BitFieldStruct):
    _FIELDNAMES = ['value']
    _BITFIELDS = ['target', 'diversity', 'addrDiv', 'key', 'next', 'bind', 'auth']
    _SIZES = [uint64_t]
    _BF_SIZES = [32, 16, 1, 2, 11, 1, 1]
    SIZE = uint64_t

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)


class dyld_chained_ptr_arm64e_auth_bind(BitFieldStruct):
    _FIELDNAMES = ['value']
    _BITFIELDS = ['ordinal', 'zero', 'diversity', 'addrDiv', 'key', 'next', 'bind', 'auth']
    _SIZES = [uint64_t]
    _BF_SIZES = [16, 16, 16, 1, 2, 11, 1, 1]
    SIZE = uint64_t

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)


class dyld_chained_ptr_64(BitFieldStruct):
    _FIELDNAMES = ['value']
    _BITFIELDS = ['reserved', 'bind']
    _SIZES = [8]
    _BF_SIZES = [63, 1]
    SIZE = 8

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)


class dyld_chained_ptr_64_rebase(BitFieldStruct):
    _FIELDNAMES = ['value']
    _BITFIELDS = ['target', 'high8', 'reserved', 'next', 'bind']
    _SIZES = [8]
    _BF_SIZES = [36, 8, 7, 12, 1]
    SIZE = 8

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)


class dyld_chained_ptr_64_bind(BitFieldStruct):
    _FIELDNAMES = ['value']
    _BITFIELDS = ['ordinal', 'addend', 'reserved', 'next', 'bind']
    _SIZES = [uint64_t]
    _BF_SIZES = [24, 8, 19, 12, 1]
    SIZE = uint64_t

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)


class dyld_chained_import(Struct):
    _FIELDNAMES = ['_import']
    _SIZES = [4]
    SIZE = 4

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)


class dyld_chained_import_addend(Struct):
    _FIELDNAMES = ['import', 'addend']
    _SIZES = [4, 4]
    SIZE = 8

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)


class dyld_chained_import_addend64(Struct):
    _FIELDNAMES = ['import', 'addend']
    _SIZES = [8, 8]
    SIZE = 16

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)


DYLD_CHAINED_PTR_BASE = {
    dyld_chained_ptr_format.DYLD_CHAINED_PTR_ARM64E: dyld_chained_ptr_arm64e,
    dyld_chained_ptr_format.DYLD_CHAINED_PTR_NOIDEA: dyld_chained_ptr_arm64e,
    dyld_chained_ptr_format.DYLD_CHAINED_PTR_64: dyld_chained_ptr_64
}




#
#  ktool | ktool
#  objc.py
#
#  This file contains utilities for parsing objective C classes within a MachO binary
#
#  This file is part of ktool. ktool is free software that
#  is made available under the MIT license. Consult the
#  file "LICENSE" that is distributed together with this file
#  for the exact licensing terms.
#
#  Copyright (c) kat 2021.
#
from collections import namedtuple
from enum import Enum
from typing import List, Dict, Optional

from kmacho.base import Constructable
from ktool.dyld import Image
from ktool.structs import *
from ktool.util import log, ignore, usi32_to_si32, opts, Queue, QueueItem

type_encodings = {
    "c": "char",
    "i": "int",
    "s": "short",
    "l": "long",
    "q": "NSInteger",
    "C": "unsigned char",
    "I": "unsigned int",
    "S": "unsigned short",
    "L": "unsigned long",
    "A": "uint8_t",
    "Q": "NSUInteger",
    "f": "float",
    "d": "CGFloat",
    "b": "BOOL",
    "@": "id",
    "B": "BOOL",
    "v": "void",
    "*": "char *",
    "#": "Class",
    ":": "SEL",
    "?": "unk",
    "T": "unk"
}

# https://github.com/arandomdev/DyldExtractor/blob/master/DyldExtractor/objc/objc_structs.py#L79
RELATIVE_METHODS_SELECTORS_ARE_DIRECT_FLAG = 0x40000000
RELATIVE_METHOD_FLAG = 0x80000000
METHOD_LIST_FLAGS_MASK = 0xFFFF0000


class ObjCImage(Constructable):
    @classmethod
    def from_image(cls, image: Image):

        objc_image = ObjCImage(image)

        cat_prot_queue = Queue()
        class_queue = Queue()

        cat_prot_queue.multithread = False
        class_queue.multithread = False

        sect = None
        for seg in image.segments:
            for sec in image.segments[seg].sections:
                if sec == "__objc_catlist":
                    sect = image.segments[seg].sections[sec]

        # cats = []  # meow
        if sect is not None:
            count = sect.size // 0x8
            for offset in range(0, count):
                try:
                    item = QueueItem()
                    item.func = Category.from_image
                    item.args = [objc_image, sect.vm_address + offset * 0x8]
                    cat_prot_queue.items.append(item)
                except Exception as ex:
                    if not ignore.OBJC_ERRORS:
                        raise ex
                    log.error(f'Failed to load a category! Ex: {str(ex)}')

        sect = None
        for seg in image.segments:
            for sec in image.segments[seg].sections:
                if sec == "__objc_classlist":
                    sect = image.segments[seg].sections[sec]

        if sect is not None:
            cnt = sect.size // 0x8
            for i in range(0, cnt):
                try:
                    # c = Class.from_image(objc_image, sect.vm_address + i * 0x8)
                    item = QueueItem()
                    item.func = Class.from_image
                    item.args = [objc_image, sect.vm_address + i * 0x8]
                    class_queue.items.append(item)
                except Exception as ex:
                    if not ignore.OBJC_ERRORS:
                        raise ex
                    log.error(f'Failed to load a class! Ex: {str(ex)}')

        sect = None
        for seg in image.segments:
            for sec in image.segments[seg].sections:
                if sec == "__objc_protolist":
                    sect = image.segments[seg].sections[sec]

        if sect is not None:
            cnt = sect.size // 0x8
            for i in range(0, cnt):
                ptr = sect.vm_address + i * 0x8
                if objc_image.vm_check(ptr):
                    loc = image.get_int_at(ptr, 0x8, vm=True)
                    try:
                        proto = image.load_struct(loc, objc2_prot, vm=True)
                        item = QueueItem()
                        item.func = Protocol.from_image
                        item.args = [objc_image, proto, loc]
                        cat_prot_queue.items.append(item)
                    except Exception as ex:
                        if not ignore.OBJC_ERRORS:
                            raise ex
                        log.error("Failed to load a protocol with " + str(ex))

        cat_prot_queue.go()

        for val in cat_prot_queue.returns:
            if val:
                if isinstance(val, Protocol):
                    objc_image.protolist.append(val)
                    objc_image.prot_map[val.loc] = val
                else:
                    objc_image.catlist.append(val)
                    objc_image.cat_map[val.loc] = val

        class_queue.go()

        for val in class_queue.returns:
            if val:
                objc_image.classlist.append(val)
                objc_image.class_map[val.loc] = val

        return objc_image

    @classmethod
    def from_values(cls, image, name, classlist, catlist, protolist, type_processor=None):
        objc_image = cls(image, type_processor)
        objc_image.name = name

        objc_image.classlist = classlist
        objc_image.catlist = catlist
        objc_image.protolist = protolist

        return objc_image

    def raw_bytes(self):
        pass

    def __init__(self, image, type_processor=None):
        if type_processor is None:
            type_processor = TypeProcessor()
        self.image = image
        if image:
            self.name = image.base_name
        else:
            self.name = ""
        self.tp = type_processor

        self.classlist = []
        self.catlist = []
        self.protolist = []

        self.class_map: Dict[int, 'Class'] = {}
        self.cat_map: Dict[int, 'Category'] = {}
        self.prot_map: Dict[int, 'Protocol'] = {}

    def serialize(self):
        return {
            'classes': [cls.serialize() for cls in self.classlist],
            'categories': [cat.serialize() for cat in self.catlist],
            'protocols': [prot.serialize() for prot in self.protolist]
        }

    def vm_check(self, address):
        return self.image.vm.vm_check(address)

    def get_int_at(self, offset: int, length: int, vm=False, sectname=None):
        return self.image.get_int_at(offset, length, vm, sectname)

    def load_struct(self, addr: int, struct_type, vm=True, sectname=None, endian="little"):
        return self.image.load_struct(addr, struct_type, vm, sectname, endian)

    def get_str_at(self, addr: int, count: int, vm=True, sectname=None):
        return self.image.get_str_at(addr, count, vm, sectname)

    def get_cstr_at(self, addr: int, limit: int = 0, vm=True, sectname=None):
        return self.image.get_cstr_at(addr, limit, vm, sectname)


class Struct_Representation:
    def __init__(self, processor: 'TypeProcessor', type_str: str):
        # {name=dd}

        # Remove the outer {}, then get everything to the left of the equal sign
        self.name: str = type_str[1:-1].split('=')[0]

        if '=' not in type_str:
            self.fields = []
            return

        self.field_names = []

        process_string = type_str[1:-1].split('=', 1)[1]

        if process_string.startswith('"'):  # Named struct
            output_string = ""

            in_field = False
            in_substruct_depth = 0

            field = ""

            for character in process_string:
                if character == '{':
                    in_substruct_depth += 1
                    output_string += character
                    continue

                elif character == '}':
                    in_substruct_depth -= 1
                    output_string += character
                    continue

                if in_substruct_depth == 0:
                    if character == '"':
                        if in_field:
                            self.field_names.append(field)
                            in_field = False
                            field = ""
                        else:
                            in_field = True
                    else:
                        if in_field:
                            field += character
                        else:
                            output_string += character
                else:
                    output_string += character

            process_string = output_string

        # Remove the outer {},
        # get everything after the first = sign,
        # Process that via the processor
        # Save the resulting list to self.fields
        self.fields = processor.process(process_string)

    def __str__(self):
        ret = "typedef struct " + self.name + " {\n"

        if not self.fields:
            ret += "} // Error Processing Struct Fields"
            return ret

        for i, field in enumerate(self.fields):
            field_name = f'field{str(i)}'

            if len(self.field_names) > 0:
                try:
                    field_name = self.field_names[i]
                except IndexError:
                    log.debug(f'Missing a field in struct {self.name}')

            if isinstance(field.value, Struct_Representation):
                field = field.value.name
            else:
                field = field.value

            ret += "    " + field + ' ' + field_name + ';\n'
        ret += '} ' + self.name + ';'
        if len(self.fields) == 0:
            ret += " // Error Processing Struct Fields"
        return ret


class EncodingType(Enum):
    METHOD = 0
    PROPERTY = 1
    IVAR = 2


class EncodedType(Enum):
    STRUCT = 0
    NAMED = 1
    ID = 2
    NORMAL = 3


class Type:
    def __init__(self, processor, type_string, pc=0):
        start = type_string[0]
        self.child = None
        self.pointer_count = pc

        if start in type_encodings.keys():
            self.type = EncodedType.NORMAL
            self.value = type_encodings[start]
            return

        elif start == '"':
            self.type = EncodedType.NAMED
            self.value = type_string[1:-1]
            return

        elif start == '{':
            self.type = EncodedType.STRUCT
            self.value = Struct_Representation(processor, type_string)
            return
        raise ValueError(f'Struct with type {start} not found')

    def __str__(self):
        pref = ""
        for i in range(0, self.pointer_count):
            pref += "*"
        return pref + str(self.value)


class TypeProcessor:
    def __init__(self):
        self.structs = {}
        self.type_cache = {}

    def save_struct(self, struct_to_save: Struct_Representation):
        if struct_to_save.name not in self.structs.keys():
            self.structs[struct_to_save.name] = struct_to_save
        else:
            if len(self.structs[struct_to_save.name].fields) == 0:
                self.structs[struct_to_save.name] = struct_to_save
            # If the struct being saved has more field names than the one we already have saved,
            #   save this one instead.
            if len(struct_to_save.field_names) > 0 and len(self.structs[struct_to_save.name].field_names) == 0:
                self.structs[struct_to_save.name] = struct_to_save

    def process(self, type_to_process: str):
        if type_to_process in self.type_cache:
            return self.type_cache[type_to_process]
        # noinspection PyBroadException
        try:
            tokens = self.tokenize(type_to_process)
            types = []
            pc = 0
            for i, token in enumerate(tokens):
                if token == "^":
                    pc += 1
                else:
                    typee = Type(self, token, pc)
                    types.append(typee)
                    if typee.type == EncodedType.STRUCT:
                        self.save_struct(typee.value)
                    pc = 0
            self.type_cache[type_to_process] = types
            return types
        except Exception:
            pass

    @staticmethod
    def tokenize(type_to_tokenize: str):
        # ^Idd^{structZero=dd{structName={innerStructName=dd}}{structName2=dd}}

        # This took way too long to write
        # Apologies for lack of readability, it splits up the string into a list
        # Makes every character a token, except root structs
        #   which it compiles into a full string with the contents and tacks onto said list
        tokens = []
        parsing_brackets = False
        bracket_count = 0
        buffer = ""
        for c in type_to_tokenize:
            if parsing_brackets:
                buffer += c
                if c == "{":
                    bracket_count += 1
                elif c == "}":
                    bracket_count -= 1
                    if bracket_count == 0:
                        tokens.append(buffer)
                        parsing_brackets = False
                        buffer = ""
            elif c in type_encodings or c == "^":
                tokens.append(c)
            elif c == "{":
                buffer += "{"
                parsing_brackets = True
                bracket_count += 1
            elif c == '"':
                try:
                    tokens = [type_to_tokenize.split('@', 1)[1]]
                except Exception as ex:
                    log.warning(f'Failed to process type {type_to_tokenize} with {ex}')
                    return []
                break
        return tokens


class Ivar(Constructable):

    @classmethod
    def from_image(cls, objc_image: ObjCImage, ivar: objc2_ivar):
        name: str = objc_image.get_cstr_at(ivar.name, 0, True, "__objc_methname")
        type_string: str = objc_image.get_cstr_at(ivar.type, 0, True, "__objc_methtype")
        return cls(name, type_string, objc_image.tp)

    @classmethod
    def from_values(cls, name, type_encoding, type_processor=None):
        if not type_processor:
            type_processor = TypeProcessor()
        return cls(name, type_encoding, type_processor)

    def raw_bytes(self):
        pass

    def __init__(self, name, type_encoding, type_processor):
        self.name: str = name
        type_string: str = type_encoding
        self.typestr = type_string
        if len(type_string) == 0:
            self.is_id = False
            self.type = '?'
            return
        self.is_id: bool = type_string[0] == "@"
        try:
            self.type: str = self._renderable_type(type_processor.process(type_string)[0])
        except IndexError:
            self.type: str = '?'

    def serialize(self):
        return {
            'name': self.name,
            'type': self.type,
            'type_is_id': self.type,
            'typestring': self.typestr,
            'rendered': str(self)
        }

    def __str__(self):
        ret = ""
        if self.type.startswith('<'):
            ret += "id"
        ret += self.type + ' '
        if self.is_id:
            ret += '*'
        ret += self.name
        return ret

    @staticmethod
    def _renderable_type(ivar_type: Type) -> str:
        if ivar_type.type == EncodedType.NORMAL:
            return str(ivar_type)
        elif ivar_type.type == EncodedType.STRUCT:
            ptr_addition = ""
            for i in range(0, ivar_type.pointer_count):
                ptr_addition += '*'
            return ptr_addition + ivar_type.value.name
        return str(ivar_type)


class MethodList:
    def __init__(self, image: ObjCImage, methlist_head, base_meths, class_meta, class_name):
        log.info(f'Opening method list ({str(methlist_head)}) at {hex(base_meths)}')
        self.objc_image = image
        self.methlist_head = methlist_head
        self.meta = class_meta
        self.name = class_name
        self.load_errors = []
        self.methods = []

        self.struct_list = []
        if base_meths != 0:
            self.methods = self._process_methlist(base_meths)

    def _process_methlist(self, base_meths):
        methods = []

        ea = self.methlist_head.off
        vm_ea = base_meths

        uses_relative_methods = self.methlist_head.entrysize & METHOD_LIST_FLAGS_MASK & RELATIVE_METHOD_FLAG != 0
        rms_are_direct = self.methlist_head.entrysize & METHOD_LIST_FLAGS_MASK & RELATIVE_METHODS_SELECTORS_ARE_DIRECT_FLAG != 0

        ea += objc2_meth_list.SIZE
        vm_ea += objc2_meth_list.SIZE

        for i in range(1, self.methlist_head.count + 1):
            if uses_relative_methods:
                sel = self.objc_image.get_int_at(ea, 4, vm=False)
                types = self.objc_image.get_int_at(ea + 4, 4, vm=False)
            else:
                sel = self.objc_image.get_int_at(ea, 8, vm=False)
                types = self.objc_image.get_int_at(ea + 8, 8, vm=False)

            try:
                method = Method.from_image(self.objc_image, sel, types, self.meta, vm_ea, uses_relative_methods,
                                           rms_are_direct)
                methods.append(method)
                for method_type in method.types:
                    if method_type.type == EncodedType.STRUCT:
                        self.struct_list.append(method_type.value)

            except Exception as ex:
                if not ignore.OBJC_ERRORS:
                    raise ex
                log.warning(f'Failed to load method in {self.name} with {str(ex)}')
                self.load_errors.append(f'Failed to load a method with {str(ex)}')

            if uses_relative_methods:
                ea += objc2_meth_list_entry.SIZE
                vm_ea += objc2_meth_list_entry.SIZE
            else:
                ea += objc2_meth.SIZE
                vm_ea += objc2_meth.SIZE

        return methods


class Method(Constructable):
    @classmethod
    def from_image(cls, objc_image: ObjCImage, sel_addr, types_addr, is_meta, vm_addr, rms, rms_are_direct):
        if rms:
            if rms_are_direct:
                try:
                    if opts.USE_SYMTAB_INSTEAD_OF_SELECTORS:
                        raise AssertionError
                    sel = objc_image.get_cstr_at(sel_addr + vm_addr, 0, vm=True, sectname="__objc_methname")

                except Exception as ex:
                    try:
                        imp = objc_image.get_int_at(vm_addr + 8, 4, vm=True)
                        imp = usi32_to_si32(imp) + vm_addr + 8
                        if imp in objc_image.image.symbols:
                            sel = objc_image.image.symbols[imp].fullname.split(" ")[-1][:-1]
                        else:
                            raise ex
                    except Exception:
                        raise ex
                type_string = objc_image.get_cstr_at(types_addr + vm_addr + 4, 0, vm=True,
                                                     sectname="__objc_methtype")
            else:
                selector_pointer = objc_image.get_int_at(sel_addr + vm_addr, 8, vm=True)
                try:
                    if opts.USE_SYMTAB_INSTEAD_OF_SELECTORS:
                        raise AssertionError
                    sel = objc_image.get_cstr_at(selector_pointer, 0, vm=True, sectname="__objc_methname")
                except Exception as ex:
                    try:
                        imp = objc_image.get_int_at(vm_addr + 8, 4, vm=True)
                        # no idea if this is correct
                        imp = usi32_to_si32(imp) + vm_addr + 8
                        if imp in objc_image.image.symbols:
                            sel = objc_image.image.symbols[imp].fullname.split(" ")[-1][:-1]
                        else:
                            raise ex
                    except Exception:
                        raise ex
                type_string = objc_image.get_cstr_at(types_addr + vm_addr + 4, 0, vm=True,
                                                     sectname="__objc_methtype")
        else:
            sel = objc_image.get_cstr_at(sel_addr, 0, vm=True, sectname="__objc_methname")
            type_string = objc_image.get_cstr_at(types_addr, 0, vm=True, sectname="__objc_methtype")
        return cls(is_meta, sel, type_string, objc_image.tp)

    @classmethod
    def from_values(cls, sel, type_string, is_meta=False, type_processor=None):
        if not type_processor:
            type_processor = TypeProcessor()
        return cls(is_meta, sel, type_string, type_processor)

    def raw_bytes(self):
        pass

    def __init__(self, meta, sel, type_string, type_processor):
        self.meta = meta
        self.sel = sel
        self.type_string = type_string
        self.types = type_processor.process(type_string)

        self.return_string = self._renderable_type(self.types[0])
        self.arguments = [self._renderable_type(i) for i in self.types[1:]]

        self.signature = self._build_method_signature()

    def serialize(self):
        return {
            'selector': self.sel,
            'arguments': self.arguments,
            'return_type': self.return_string,
            'signature': self.signature,
            'typestring': self.type_string
        }

    def __str__(self):
        ret = ""
        ret += self.signature
        return ret

    @staticmethod
    def _renderable_type(method_type: Type):
        if method_type.type == EncodedType.NORMAL:
            return str(method_type)
        elif method_type.type == EncodedType.STRUCT:
            ptr_addition = ""
            for i in range(0, method_type.pointer_count):
                ptr_addition += '*'
            return 'struct ' + method_type.value.name + ' ' + ptr_addition

    def _build_method_signature(self):
        dash = "+" if self.meta else "-"
        ret = "(" + self.return_string + ")"

        if len(self.arguments) == 0:
            return dash + ret + self.sel

        segments = []
        for i, item in enumerate(self.sel.split(':')):
            if item == "":
                continue
            try:
                segments.append(item + ':' + '(' + self.arguments[i + 2] + ')' + 'arg' + str(i) + ' ')
            except IndexError:
                segments.append(item)

        sig = ''.join(segments)

        return dash + ret + sig


class LinkedClass:
    def __init__(self, classname, libname):
        self.classname = classname
        self.libname = libname


class Class(Constructable):
    """
    """

    @classmethod
    def from_image(cls, objc_image: ObjCImage, class_ptr: int, meta=False) -> Optional['Class']:
        if class_ptr in objc_image.class_map:
            return objc_image.class_map[class_ptr]

        load_errors = []
        struct_list = []

        if not objc_image.vm_check(class_ptr):
            objc2_class_location = objc_image.get_int_at(class_ptr, 8, vm=False)
        else:
            objc2_class_location = objc_image.get_int_at(class_ptr, 8, vm=True)

        if objc2_class_location == 0 or not objc_image.vm_check(objc2_class_location):
            log.error("Loading a class failed")
            return None

        objc2_class_item: objc2_class = objc_image.load_struct(objc2_class_location, objc2_class, vm=True)

        superclass = None

        if not meta:
            if objc2_class_location + 8 in objc_image.image.import_table:
                symbol = objc_image.image.import_table[objc2_class_location + 8]
                superclass_name = symbol.name[1:]
            elif objc2_class_item.superclass in objc_image.image.export_table:
                symbol = objc_image.image.export_table[objc2_class_item.superclass]
                superclass_name = symbol.name[1:]
            else:
                if objc_image.vm_check(objc2_class_item.superclass):
                    # noinspection PyBroadException
                    try:
                        superclass = Class.from_image(objc_image, objc2_class_location + 8)
                    except Exception:
                        pass
                if superclass is not None:
                    superclass_name = superclass.name
                else:
                    if objc2_class_item.superclass in objc_image.image.import_table:
                        symbol = objc_image.image.import_table[objc2_class_item.superclass]
                        superclass_name = symbol.name[1:]
                    else:
                        superclass_name = 'NSObject'
        else:
            superclass_name = ''

        ro_location = objc2_class_item.info >> (1 << 1) << 2

        try:
            objc2_class_ro_item = objc_image.load_struct(ro_location, objc2_class_ro, vm=True)
        except ValueError:
            log.warn("Class Data is off-image")
            return None
        if not meta:
            try:
                name = objc_image.get_cstr_at(objc2_class_ro_item.name, 0, vm=True)
            except ValueError:
                log.warning(f'Classname out of bounds')
                name = ""
        else:
            name = ""

        methods = []
        properties = []

        if objc2_class_ro_item.base_props != 0:
            proplist_head = objc_image.load_struct(objc2_class_ro_item.base_props, objc2_prop_list)
            ea = proplist_head.off
            ea += objc2_prop_list.SIZE

            for i in range(1, proplist_head.count + 1):
                prop = objc_image.load_struct(ea, objc2_prop, vm=False)

                try:
                    property = Property.from_image(objc_image, prop)
                    properties.append(property)
                    if hasattr(property, 'attr'):
                        if property.attr.type.type == EncodedType.STRUCT:
                            struct_list.append(property.attr.type.value)

                except Exception as ex:
                    if not ignore.OBJC_ERRORS:
                        raise ex
                    log.warning(f'Failed to load a property in {name} with {ex.__class__.__name__}: {str(ex)}')
                    load_errors.append(f'Failed to load a property with {ex.__class__.__name__}: {str(ex)}')

                ea += objc2_prop.SIZE

        if objc2_class_ro_item.base_meths != 0:
            methlist_head = objc_image.load_struct(objc2_class_ro_item.base_meths, objc2_meth_list)

            methlist = MethodList(objc_image, methlist_head, objc2_class_ro_item.base_meths, meta, name)

            load_errors += methlist.load_errors
            struct_list += methlist.struct_list
            methods += methlist.methods

        if objc2_class_item.isa != 0 and objc2_class_item.isa <= 0xFFFFFFFFFF and not meta:
            metaclass = Class.from_image(objc_image, objc2_class_item.off, meta=True)
            if metaclass:
                methods += metaclass.methods

        prots = []
        if objc2_class_ro_item.base_prots != 0:
            protlist: objc2_prot_list = objc_image.load_struct(objc2_class_ro_item.base_prots, objc2_prot_list)
            ea = protlist.off
            for i in range(1, protlist.cnt + 1):
                prot_loc = objc_image.get_int_at(ea + i * 8, 8, vm=False)
                if prot_loc in objc_image.prot_map:
                    prots.append(objc_image.prot_map[prot_loc])
                else:
                    prot = objc_image.load_struct(prot_loc, objc2_prot, vm=True)
                    try:
                        p = Protocol.from_image(objc_image, prot, prot_loc)
                        prots.append(p)
                        objc_image.prot_map[prot_loc] = p
                    except Exception as ex:
                        if not ignore.OBJC_ERRORS:
                            raise ex
                        log.warning(f'Failed to load protocol with {str(ex)}')
                        load_errors.append(f'Failed to load a protocol with {str(ex)}')

        ivars = []
        if objc2_class_ro_item.ivars != 0:
            ivarlist: objc2_ivar_list = objc_image.load_struct(objc2_class_ro_item.ivars, objc2_ivar_list)
            ea = ivarlist.off + 8
            for i in range(1, ivarlist.cnt + 1):
                ivar_loc = ea + objc2_ivar.SIZE * (i - 1)
                ivar = objc_image.load_struct(ivar_loc, objc2_ivar, vm=False)
                try:
                    ivar_object = Ivar.from_image(objc_image, ivar)
                    ivars.append(ivar_object)
                except Exception as ex:
                    if not ignore.OBJC_ERRORS:
                        raise ex
                    log.warning(f'Failed to load ivar with {str(ex)}')
                    load_errors.append(f'Failed to load an ivar with {str(ex)}')

        return cls(name, meta, superclass_name, methods, properties, ivars, prots, load_errors, struct_list,
                   loc=objc2_class_location)

    @classmethod
    def from_values(cls, name, superclass_name, methods: List[Method], properties: List['Property'],
                    ivars: List['Ivar'],
                    protocols: List['Protocol'], load_errors=None, structs=None):
        return cls(name, False, superclass_name, methods, properties, ivars, protocols, load_errors, structs)

    def raw_bytes(self):
        pass

    def __init__(self, name, is_meta, superclass_name, methods, properties, ivars, protocols, load_errors=None,
                 structs=None, loc=0):
        if structs is None:
            structs = []
        if load_errors is None:
            load_errors = []
        self.name = name
        self.meta = is_meta
        self.superclass = superclass_name
        self.loc = loc

        self.load_errors = load_errors
        self.struct_list = structs

        self.linkedlibs = []
        self.linked_classes = []
        self.fdec_classes = []
        self.fdec_prots = []

        self.methods = methods
        self.properties = properties
        self.protocols = protocols
        self.ivars = ivars

    def serialize(self):
        return {
            'name': self.name,
            'superclass': self.superclass,
            'methods': [meth.serialize() for meth in self.methods],
            'properties': [prop.serialize() for prop in self.properties],
            'protocols': [prot.name for prot in self.protocols],
            'ivars': [ivar.serialize() for ivar in self.ivars]
        }

    def __str__(self):
        ret = ""
        ret += self.name
        return ret

    def _load_linked_libraries(self):
        pass


attr_encodings = {
    "&": "retain",
    "N": "nonatomic",
    "W": "__weak",
    "R": "readonly",
    "C": "copy"
}

property_attr = namedtuple("property_attr", ["type", "attributes", "ivar", "is_id", "typestr"])


class Property(Constructable):

    @classmethod
    def from_image(cls, objc_image: ObjCImage, property: objc2_prop):
        name = objc_image.get_cstr_at(property.name, 0, True, "__objc_methname")
        attr_string = objc_image.get_cstr_at(property.attr, 0, True, "__objc_methname")
        return cls(name, attr_string, objc_image.tp)

    @classmethod
    def from_values(cls, name, attr_string, type_processor=None):
        if not type_processor:
            type_processor = TypeProcessor()
        return cls(name, attr_string, type_processor)

    def raw_bytes(self):
        pass

    def __init__(self, name, attr_string, type_processor):
        self.name = name

        try:
            self.attr = self.decode_property_attributes(type_processor, attr_string)
        except IndexError:
            log.warn(
                f'issue with property {self.name} attr {attr_string}')
            self.type = None
            self.is_id = False
            self.attributes = []
            self.ivarname = ""
            self.attr = None

        self.type = self._renderable_type(self.attr.type)
        self.is_id = self.attr.is_id
        self.attributes = self.attr.attributes
        self.ivarname = self.attr.ivar

    def serialize(self):
        return {
            'name': self.name,
            'type': self.type,
            'is_id': self.is_id,
            'ivar_name': self.ivarname,
            'attributes': self.attributes,
            'rendered': str(self)
        }

    def __str__(self):
        if not hasattr(self, 'attributes'):
            return f'// Something went wrong loading property {self.name}'
        ret = "@property "

        if len(self.attributes) > 0:
            ret += '(' + ', '.join(self.attributes) + ') '

        if self.type.startswith('<'):
            ret += "id"
        ret += self.type + ' '

        if self.is_id:
            ret += '*'

        ret += self.name
        return ret

    @staticmethod
    def _renderable_type(_type: Type):
        if _type.type == EncodedType.NORMAL:
            return str(_type)
        elif _type.type == EncodedType.STRUCT:
            ptraddon = ""
            for i in range(0, _type.pointer_count):
                ptraddon += '*'
            return ptraddon + _type.value.name
        return str(_type)

    @staticmethod
    def decode_property_attributes(type_processor, type_str: str):
        attribute_strings = type_str.split(',')

        ptype = ""
        is_id = False
        ivar = ""
        property_attributes = []

        for attribute in attribute_strings:
            indicator = attribute[0]
            if indicator == "T":
                ptype = type_processor.process(attribute[1:])[0]
                if ptype == "{":
                    print(attribute)
                is_id = attribute[1] == "@"
                continue
            if indicator == "V":
                ivar = attribute[1:]
            if indicator in attr_encodings:
                property_attributes.append(attr_encodings[indicator])

        return property_attr(ptype, property_attributes, ivar, is_id, type_str)


class Category(Constructable):

    @classmethod
    def from_image(cls, objc_image: ObjCImage, category_ptr):
        try:
            loc = objc_image.get_int_at(category_ptr, 8, vm=True)
            struct: objc2_category = objc_image.load_struct(loc, objc2_category, vm=True)
            name = objc_image.get_cstr_at(struct.name, vm=True)
        except ValueError as ex:
            log.error("Couldn't load basic info about a Category: " + str(ex))
            return None

        classname = ""
        try:
            sym = objc_image.image.import_table[loc + 8]
            classname = sym.name[1:]
        except KeyError:
            pass

        methods = []
        properties = []
        load_errors = []
        struct_list = []

        if struct.inst_meths != 0:
            try:
                methlist_head = objc_image.load_struct(struct.inst_meths, objc2_meth_list)
                methlist = MethodList(objc_image, methlist_head, struct.inst_meths, False, f'{classname}+{name}')

                load_errors += methlist.load_errors
                struct_list += methlist.struct_list
                methods += methlist.methods
            except ValueError:
                log.warn("Methods for this category are off-image.")

        if struct.class_meths != 0:
            try:
                methlist_head = objc_image.load_struct(struct.class_meths, objc2_meth_list)
                methlist = MethodList(objc_image, methlist_head, struct.class_meths, True, f'{classname}+{name}')

                load_errors += methlist.load_errors
                struct_list += methlist.struct_list
                methods += methlist.methods
            except ValueError:
                log.warn("Methods for this category are off-image.")

        if struct.props != 0:
            try:
                proplist_head = objc_image.load_struct(struct.props, objc2_prop_list)

                ea = proplist_head.off
                ea += 8

                for i in range(1, proplist_head.count + 1):
                    prop = objc_image.load_struct(ea, objc2_prop, vm=False)
                    try:
                        properties.append(Property.from_image(objc_image, prop))
                    except Exception as ex:
                        log.warning(f'Failed to load property with {str(ex)}')
                    ea += objc2_prop.SIZE

            except ValueError:
                log.warn("Properties for this category are off-image.")

        return cls(classname, name, methods, properties, loc=loc)

    @classmethod
    def from_values(cls, classname, name, methods, properties, load_errors=None, struct_list=None):
        return cls(classname, name, methods, properties, load_errors, struct_list)

    def raw_bytes(self):
        pass

    def __init__(self, classname, name, methods, properties, load_errors=None, struct_list=None, loc=0):
        if load_errors is None:
            load_errors = []
        if struct_list is None:
            struct_list = []
        self.name = name
        self.classname = classname
        self.loc = loc

        self.load_errors = load_errors
        self.struct_list = struct_list

        self.methods = methods
        self.properties = properties
        self.protocols = []

    def serialize(self):
        return {
            'name': self.name,
            'classname': self.classname,
            'methods': [method.serialize() for method in self.methods],
            'properties': [prop.serialize() for prop in self.properties],
            'protocols': [prot.name for prot in self.protocols]
        }


class Protocol(Constructable):

    @classmethod
    def from_image(cls, objc_image: 'ObjCImage', protocol: objc2_prot, loc):
        if loc in objc_image.prot_map:
            return objc_image.prot_map[loc]

        name = objc_image.get_cstr_at(protocol.name, 0, vm=True)
        load_errors = []
        struct_list = []

        methods = []
        opt_methods = []

        properties = []

        methlist = Protocol.load_methods(objc_image, name, protocol.inst_meths)
        load_errors += methlist.load_errors
        struct_list += methlist.struct_list
        methods += methlist.methods

        methlist = Protocol.load_methods(objc_image, name, protocol.class_meths, True)
        load_errors += methlist.load_errors
        struct_list += methlist.struct_list
        methods += methlist.methods

        methlist = Protocol.load_methods(objc_image, name, protocol.opt_inst_meths)
        load_errors += methlist.load_errors
        struct_list += methlist.struct_list
        opt_methods += methlist.methods

        methlist = Protocol.load_methods(objc_image, name, protocol.opt_class_meths, True)
        load_errors += methlist.load_errors
        struct_list += methlist.struct_list
        opt_methods += methlist.methods

        if protocol.inst_props != 0:
            proplist_head = objc_image.load_struct(protocol.inst_props, objc2_prop_list)

            ea = proplist_head.off
            ea += 8

            for i in range(1, proplist_head.count + 1):
                prop = objc_image.load_struct(ea, objc2_prop, vm=False)
                try:
                    properties.append(Property.from_image(objc_image, prop))
                except Exception as ex:
                    log.warning(f'Failed to load property with {str(ex)}')
                ea += objc2_prop.SIZE

        return cls(name, methods, opt_methods, properties, load_errors, struct_list, loc=loc)

    @classmethod
    def load_methods(cls, objc_image, name, loc, meta=False):
        vm_ea = loc
        if loc != 0:
            methlist_head = objc_image.load_struct(loc, objc2_meth_list)
        else:
            methlist_head = None

        methlist = MethodList(objc_image, methlist_head, vm_ea, meta, name)

        return methlist

    @classmethod
    def from_values(cls, name, methods, opt_methods, properties, load_errors=None, struct_list=None):
        return cls(name, methods, opt_methods, properties, load_errors, struct_list)

    def raw_bytes(self):
        pass

    def __init__(self, name, methods, opt_methods, properties, load_errors=None, struct_list=None, loc=0):
        if struct_list is None:
            struct_list = []
        if load_errors is None:
            load_errors = []

        self.name = name
        self.loc = loc

        self.load_errors = load_errors
        self.struct_list = struct_list

        self.methods = methods

        self.opt_methods = opt_methods

        self.properties = properties

    def serialize(self):
        return {
            'name': self.name,
            'methods': [meth.serialize() for meth in self.methods],
            'optional-methods': [meth.serialize() for meth in self.opt_methods],
            'properties': [prop.serialize() for prop in self.properties]
        }

    def __str__(self):
        return self.name
